import unittest
from pathlib import Path
from coldtype.geometry import *
from coldtype.color import hsl
from coldtype.pens.datpen import DPS
from coldtype.text.composer import StSt, Font, Style

tf = Path(__file__).parent

class TestText(unittest.TestCase):
    def _test_glyph_names(self, r, font_path):
        ss = StSt("CDELOPTY", font_path, 300, wdth=0).align(r)
        self.assertEqual(len(ss), 8)
        self.assertEqual(ss[0].glyphName, "C")
        self.assertEqual(ss[-1].glyphName, "Y")
        return ss

    def test_format_equality(self):
        r = Rect(800, 300)

        ttf = self._test_glyph_names(r, Font.ColdtypeObviously())
        otf = self._test_glyph_names(r, "assets/ColdtypeObviously_CompressedBlackItalic.otf")
        ufo = self._test_glyph_names(r, "assets/ColdtypeObviously_CompressedBlackItalic.ufo")
        ds = self._test_glyph_names(r, "assets/ColdtypeObviously.designspace")

        # TODO why isn't the ttf version equal to these?
        self.assertEqual(ufo[0].value, ds[0].value)
        self.assertEqual(ufo[-1].value, ds[-1].value)

        DPS([
            ttf, otf, ufo, ds
        ]).f(None).s(hsl(0.5, a=0.3)).sw(1).picklejar(r)
    
    def test_space(self):
        r = Rect(2000, 1000)

        txt = StSt("A B", Font.MutatorSans(), 1000)
        space = txt[1]
        self.assertEqual(space.glyphName, "space")
        self.assertEqual(space.ambit().w, 250)
        self.assertEqual(txt.ambit().w, 1093)
        self.assertEqual(space.ambit().x, 400)
        txt.align(r)
        self.assertEqual(space.ambit().x, 863.5)
        #txt.picklejar(r)

        txt = StSt("A B", Font.MutatorSans(), 1000, space=500)
        space = txt[1]
        self.assertEqual(space.glyphName, "space")
        self.assertEqual(space.ambit().w, 500)
        self.assertEqual(txt.ambit().w, 1093+250)
        self.assertEqual(space.ambit().x, 400)
        txt.align(r)
        self.assertEqual(space.ambit().x, 863.5-(250/2))
        #txt.picklejar(r)

    def test_static_fonts(self):
        r = Rect(1200, 300)
        f1 = Font.ColdtypeObviously()
        co = (StSt("CDELOPTY", f1, 200)
            .align(r)
            .picklejar(r))
        
        self.assertEqual(co.ambit(th=0).w, 998.8)
        self.assertAlmostEqual(co.ambit(th=1).w, 1018.8, 1)
        self.assertEqual(len(co), 8)
        self.assertEqual(f1.font.fontPath.stem, "ColdtypeObviously-VF")
    
    def test_color_font(self):
        txt = StSt("C", "~/Type/fonts/fonts/PappardelleParty-VF.ttf", 100, palette=2)
        self.assertEqual(len(txt), 1)
        self.assertEqual(len(txt[0]), 3)
        self.assertEqual(txt[0][0].glyphName, "C_layer_0")
        self.assertEqual(txt[0][-1].glyphName, "C_layer_2")
        self.assertAlmostEqual(txt[0][0].f().h, 176.666, 2)

        txt = StSt("C", "~/Type/fonts/fonts/PappardelleParty-VF.ttf", 100)
        self.assertAlmostEqual(txt[0][0].f().h, 196.318, 2)
    
    def test_narrowing_family(self):
        r = Rect(1080, 300)
        style = Style("~/Type/fonts/fonts/nikolai/Nikolai-Bold.otf", 200, narrower=Style("~/Type/fonts/fonts/nikolai/Nikolai-NarrowBold.otf", 200, narrower=Style("~/Type/fonts/fonts/nikolai/Nikolai-CondBold.otf", 200)))

        txt = StSt("Narrowing", style, fit=r.w)
        self.assertEqual(int(txt.ambit().w), 928)

        txt = StSt("Narrowing", style, fit=r.w-200)
        self.assertEqual(int(txt.ambit().w), 840)

        txt = StSt("Narrowing", style, fit=r.w-400)
        self.assertEqual(int(txt.ambit().w), 733)

        txt = StSt("Narrowing", style, fit=r.w-600)
        self.assertEqual(int(txt.ambit().w), 733)


if __name__ == "__main__":
    unittest.main()


import json
from pathlib import Path

def remote(command):
    (Path("~/.coldtype/command.json")
        .expanduser()
        .write_text(json.dumps(dict(action=command if isinstance(command, str) else command.value))))

def show_picklejar():
    remote("reload_source")

if __name__ == "__main__":
    remote("play_preview")


import time, threading, sys

from coldtype.renderer.config import ColdtypeConfig
from coldtype.renderer.winman.passthrough import WinmanPassthrough
from coldtype.renderer.winman.glfwskia import glfw, skia, WinmanGLFWSkia, WinmanGLFWSkiaBackground
from coldtype.renderer.winman.midi import MIDIWatcher, rtmidi
from coldtype.renderer.winman.webview import WinmanWebview
from coldtype.renderer.winman.websocket import WinmanWebsocket
from coldtype.renderer.winman.blender import WinmanBlender
from coldtype.renderable import Action


last_line = ''
new_line_event = threading.Event()

def monitor_stdin():
    # https://stackoverflow.com/questions/27174736/how-to-read-most-recent-line-from-stdin-in-python
    global last_line
    global new_line_event

    def keep_last_line():
        global last_line, new_line_event
        for line in sys.stdin:
            last_line = line
            new_line_event.set()

    keep_last_line_thread = threading.Thread(target=keep_last_line)
    keep_last_line_thread.daemon = True
    keep_last_line_thread.start()


class Winmans():
    def __init__(self, renderer, config:ColdtypeConfig):
        self.renderer = renderer
        self.config = config

        self.pt = WinmanPassthrough()

        self.glsk:WinmanGLFWSkia = None
        self.ws:WinmanWebsocket = None
        self.wv:WinmanWebview = None
        self.midi:MIDIWatcher = None
        self.b3d:WinmanBlender = None

        self.last_time = -1
        self.refresh_delay = self.config.refresh_delay
        self.backoff_refresh_delay = self.refresh_delay

        self.playing = 0
        self.preloaded_frames = []
        self.playing_preloaded_frame = -1

        self.bg = False
        if (config.args.is_subprocess
            or config.args.all
            or config.args.release
            or config.args.build
            ):
            self.bg = True

    def should_glfwskia(self):
        return glfw is not None and skia is not None and not self.config.no_viewer
    
    def should_webviewer(self):
        return self.config.webviewer
    
    def should_midi(self):
        return rtmidi and not self.config.no_midi
    
    def should_blender(self):
        return self.config.blender_watch
    
    def add_viewers(self):
        if not self.bg:
            monitor_stdin()

        if self.config.websocket or self.should_webviewer():
            self.ws = WinmanWebsocket(self.config, self.renderer)

        if self.should_glfwskia():
            self.glsk = WinmanGLFWSkia(self.config, self.renderer)
        
        if self.should_webviewer():
            self.wv = WinmanWebview(self.config, self.renderer)
        
        if self.should_blender():
            self.b3d = WinmanBlender()
        
        if self.should_midi():
            self.midi = MIDIWatcher(self.config,
                self.renderer.state,
                self.renderer.execute_string_as_shortcut_or_action)
        elif self.config.args.midi_info:
            print(">>> pip install rtmidi")

    def did_reload(self, filepath):
        if self.b3d:
            self.b3d.reload(filepath)
    
    def found_blend_files(self, blend_files):
        if len(blend_files) > 0:
            if self.b3d:
                self.b3d.launch(blend_files[0])
    
    def all(self):
        return [self.pt, self.glsk, self.wv, self.b3d]
    
    def map(self):
        for wm in self.all():
            if wm:
                yield wm
    
    def set_title(self, text):
        [wm.set_title(text) for wm in self.map()]
    
    def reset(self):
        [wm.reset() for wm in self.map()]
    
    def terminate(self):
        [wm.terminate() for wm in self.map()]
    
    def toggle_playback(self):
        if self.playing == 0:
            self.playing = 1
        else:
            self.playing = 0
        
        if not self.glsk:
            if self.b3d:
                self.b3d.toggle_playback(self.playing)
    
    def frame_offset(self, offset):
        self.renderer.state.adjust_all_frame_offsets(offset)

        if not self.glsk:
            if self.b3d:
                self.b3d.frame_offset(offset)
    
    def preload_frames(self, passes):
        for rp in passes:
            self.preloaded_frames.append(rp.output_path)
        self.playing_preloaded_frame = 0
    
    def stop_playing_preloaded(self):
        self.playing_preloaded_frame = -1
    
    def toggle_play_preloaded(self):
        if self.playing_preloaded_frame >= 0:
            self.playing_preloaded_frame = -1
            self.preloaded_frames = []
        else:
            anm = self.renderer.animation()
            passes = anm.passes(Action.RenderAll, self.renderer.state, anm.all_frames())
            self.preload_frames(passes)
    
    def should_close(self):
        return any([wm.should_close() for wm in self.map()])
    
    def send_to_external(self, action, **kwargs):
        if self.ws:
            self.ws.send_to_external(action, **kwargs)
    
    def poll(self):
        if self.glsk:
            self.glsk.poll()
    
    def turn_over(self):
        if self.midi:
            if self.midi.monitor(self.playing):
                self.renderer.action_waiting = Action.PreviewStoryboard

        if self.ws:
            self.ws.read_messages()

        render_previews = len(self.renderer.previews_waiting_to_paint) > 0
        if not render_previews:
            self.backoff_refresh_delay += 0.01
            if self.backoff_refresh_delay >= 0.25:
                self.backoff_refresh_delay = 0.25
            return []
        
        self.backoff_refresh_delay = self.refresh_delay

        did_preview = []
        if self.glsk:
            did_preview.append(self.glsk.turn_over())
        
        if self.wv:
            did_preview.append(self.wv.turn_over())
        
        return did_preview
    
    def run_loop(self):
        while (not self.renderer.dead
            and not self.should_close()
            ):
            t2 = time.time()
            td2 = t2 - self.last_time

            spf = 0.1
            if self.renderer.last_animation:
                spf = 1 / float(self.renderer.last_animation.timeline.fps)

                if td2 >= spf:
                    self.last_time = t2
                else:
                    self.poll()
                    continue
            
            if self.renderer.last_animation and self.playing_preloaded_frame >= 0 and len(self.preloaded_frames) > 0:
                if self.glsk:
                    self.glsk.show_preloaded_frame(self.preloaded_frames[self.playing_preloaded_frame])
                
                self.playing_preloaded_frame += 1
                if self.playing_preloaded_frame == len(self.preloaded_frames):
                    self.playing_preloaded_frame = 0
                
                time.sleep(0.01)
            
            else:
                time.sleep(self.backoff_refresh_delay)
                self.last_time = t2
                
                # TODO the main turn_over, why is it like this?
                self.last_previews = self.renderer.turn_over()

                global last_line
                if last_line:
                    lls = last_line.strip()
                    if lls:
                        self.renderer.on_stdin(lls)
                    last_line = None
            
            self.poll()
    
        self.renderer.on_exit(restart=False)


import unittest
from coldtype.time.nle.ascii import AsciiTimeline
from coldtype.renderable.animation import animation
from coldtype.renderer.reader import SourceReader

at = AsciiTimeline(1, """
                                    |
[a          ]
        [b          ]
                [c          ]
     ]                  [d          
""")

class TestTime(unittest.TestCase):
    def test_ascii_timeline(self):
        self.assertEqual(at.duration, 36)
        self.assertEqual(at[0].start, 0)
        self.assertEqual(at[1].end, 21)
        self.assertEqual(at[2].start, 16)
        self.assertEqual(at[-1].end, 41)
        self.assertEqual(len(at.clips), 4)
    
    def test_animation(self):
        src = "test/visuals/test_animation.py"
        sr = SourceReader(src).unlink()
        anim:animation = sr.renderables()[0]
        
        self.assertEqual(anim.duration, 26)
        self.assertEqual(anim.t.find_workarea(), [0, 1, 2])
        
        for i in range(0, anim.duration):
            p = anim.passes(None, None, indices=[i])[0]
            res = anim.run_normal(p)
            gn = chr(65+i)
            if gn == "I":
                gn = "I.narrow"
            self.assertEqual(res[0][0].glyphName, gn)
            if i == 13:
                self.assertAlmostEqual(res[0][0].f().h/360, 0)

if __name__ == "__main__":
    unittest.main()


import unittest
from coldtype.pens.skiapen import SkiaPen

from pathlib import Path
from coldtype.color import hsl
from coldtype.geometry import Rect
from coldtype.text.composer import StSt, Font
from coldtype.pens.datpen import DATPen, DATPens

from PIL import Image
import imagehash
import contextlib

co = Font.Cacheable("assets/ColdtypeObviously-VF.ttf")

renders = Path("test/renders/skia")
renders.mkdir(parents=True, exist_ok=True)

def hash_img(path):
    if path.exists():
        return (
            imagehash.colorhash(Image.open(path)), 
            imagehash.average_hash(Image.open(path)))
    else:
        return -1

@contextlib.contextmanager
def test_image(test:unittest.TestCase, path, rect=Rect(1000, 500)):
    img = (renders / path)
    hash_before = hash_img(img)
    if img.exists():
        img.unlink()

    yield(img, rect)
    
    hash_after = hash_img(img)
    test.assertEqual(hash_after, hash_before)
    test.assertEqual(img.exists(), True)

class TestCairoPen(unittest.TestCase):
    def test_skia_png(self):
        with test_image(self, "test_skia.png") as (i, r):
            dp = ((ß:=DATPens())
                .define(
                    r=r,
                    nx=100,
                    a="$rIX100SY+200")
                .gs("$a↙ $a↑|$a↖OX+$nx|65 $a↘|$a↗OX-$nx|65 ɜ")
                .f(None).s(0).sw(4)
                .append(StSt("Coldtype Cdelopty".upper(),
                    co, 100, wdth=0.5)
                    .pens()
                    .distribute_on_path(ß[0], center=-5)
                    .f(hsl(0.5)))
                .align(r))
            
            SkiaPen.Precompose(dp, r, disk=str(i))
            self.assertEqual(len(dp), 2)
            self.assertEqual(type(dp), DATPens)
    
if __name__ == "__main__":
    unittest.main()


import json

from pathlib import Path
from base64 import b64encode
from IPython.display import display, SVG, HTML
from coldtype.renderable.animation import aframe, animation, Action, Timeline
from coldtype.pens.svgpen import SVGPen
from coldtype.geometry import Rect
from coldtype.renderable.animation import FFMPEGExport
from subprocess import run
from shutil import rmtree


try:
    from coldtype.fx.skia import precompose
    from coldtype.pens.skiapen import SkiaPen
    from io import BytesIO
except ImportError:
    precompose = None


def show(fmt=None, rect=None, align=False, padding=[60, 50], th=0, tv=0, scale=0.5):
    if not precompose and fmt == "png":
        raise Exception("pip install skia-python")
    
    def _display(pen):
        nonlocal rect, fmt

        if fmt is None:
            img = pen.img()
            if img and img.get("src"):
                fmt = "png"
            else:
                fmt = "svg"

        if align and rect is not None:
            pen.align(rect)
        if rect is None:
            amb = pen.ambit(th=th, tv=tv)
            rect = Rect(amb.w+padding[0], amb.h+padding[1])
            pen.align(rect)
        
        if fmt == "png":
            src = pen.ch(precompose(rect)).img().get("src")
            with BytesIO(src.encodeToData()) as f:
                f.seek(0)  # necessary?
                b64 = b64encode(f.read()).decode("utf-8")
                display(HTML(f"<img width={rect.w*scale} src='data:image/png;base64,{b64}'/>"))
        elif fmt == "svg":
            svg = SVGPen.Composite(pen, rect, viewBox=False)
            display(SVG(svg))
        return pen
    
    return _display


def showpng(rect=None, align=False, padding=[60, 50], th=0, tv=0, scale=0.5):
    return show("png", rect, align, padding, th, tv, scale)

def showlocalpng(rect, src, scale=0.5):
    with open(src, "rb") as img_file:
        encoded_string = b64encode(img_file.read()).decode("utf-8")
        display(HTML(f"<img width={rect.w*scale} src='data:image/png;base64,{encoded_string}'/>"))

def show_frame(a, idx, scale=0.5):
    rp = a.passes(Action.PreviewIndices, None, [idx])[0]
    res = a.run_normal(rp)
    show(a.fmt, a.rect, padding=[0, 0], scale=scale)(res)
    #rp.output_path.parent.mkdir(parents=True, exist_ok=True)
    #SkiaPen.Composite(res, a.rect, str(rp.output_path))
    #showlocalpng(a.rect, rp.output_path)


js = """
function animate(name, start_playing) {
    svg = document.querySelector(`#${name} svg`);
    svg_frames = [].slice.apply(svg.querySelectorAll(".frame"));
    svg_frames.forEach((sf) => sf.style.display = "none");
    svg_frames[0].style.display = "block";

    var fps = parseFloat(svg.dataset.fps);
    var duration = parseInt(svg.dataset.duration);
    var fpsInterval = 1000 / fps;
    var then = Date.now();
    var startTime = then;
    var elapsed = 0;
    var i = 0;
    var playing = !!start_playing;

    function _animate() {
        if (!playing) {
            return;
        }
        requestAnimationFrame(_animate);
        var now = Date.now();
        elapsed = now - then;
        if (elapsed > fpsInterval) {
            then = now - (elapsed % fpsInterval);
            svg_frames.forEach((sf) => sf.style.display = "none");
            svg_frames[i%duration].style.display = "block";
            i++;
        }
    }

    svg.addEventListener("click", function() {
        if (playing) {
            playing = false;
        } else {
            playing = true;
            _animate();
        }
    });

    if (playing) {
        _animate();
    }
}
"""

def show_animation(a:animation, start=False):
    idxs = range(0, a.duration+1)
    passes = a.passes(Action.PreviewIndices, None, idxs)
    results = [a.run_normal(rp) for rp in passes]
    svg = SVGPen.Animation(results, a.rect, a.timeline.fps)
    html = f"<div id='{a.name}'>{svg}</div>"
    js_call = f"<script type='text/javascript'>{js}; animate('{a.name}', {int(start)})</script>";
    display(HTML(html + js_call), display_id=a.name)

def render_animation(a, show=[], preview_scale=0.5, scale=1):
    from tqdm.notebook import tqdm

    idxs = list(range(0, a.duration))
    passes = a.passes(Action.PreviewIndices, None, idxs)
    output_dir = passes[0].output_path.parent
    if output_dir.exists():
        rmtree(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    for idx, rp in enumerate(tqdm(passes, leave=False)):
        res = a.run_normal(rp)
        if a.fmt == "png":
            SkiaPen.Composite(res, a.rect, str(rp.output_path), scale=scale)
            if show == "*" or idx in show:
                showlocalpng(a.rect, rp.output_path, scale=preview_scale)
        elif a.fmt == "svg":
            SkiaPen.SVG(res, a.rect, str(rp.output_path), scale=scale)

def show_video(a, loops=1, verbose=False, download=False, scale=0.5):
    ffex = FFMPEGExport(a, loops=loops)
    ffex.h264()
    ffex.write(verbose=verbose)
    compressed_path = str(ffex.output_path.absolute())
    mp4 = open(compressed_path, 'rb').read()
    data_url = "data:video/mp4;base64," + b64encode(mp4).decode()
    display(HTML(f"""
    <video width={a.rect.w*scale} controls loop=true autoplay>
        <source src="%s" type="video/mp4">
    </video>
    """ % data_url))
    if download:
        try:
            from google.colab import files
            files.download(ffex.output_path)
        except ImportError:
            print("download= arg is for colab")
            pass


class notebook_animation(animation):
    def __init__(self,
        rect=(540, 540),
        preview=[0],
        interactive=True,
        preview_scale=0.5,
        render_bg=True,
        vars={},
        **kwargs
        ):
        self._preview = preview
        self._interactive = interactive
        self.preview_scale = preview_scale
        self.vars = vars

        super().__init__(rect, render_bg=render_bg, **kwargs)
    
    def __call__(self, func):
        res = super().__call__(func)

        self._interaction_file = Path(self.name + "_tmp_state.json")
        self._interaction_state = {}

        if self._preview:
            if self._interactive:
                if not self._interaction_file.exists():
                    self._interaction_file.write_text("{}")
                self.interactive_preview(self._preview[0])
            else:
                self.preview(*self._preview)
        return res
    
    def interactive_preview(self, start):
        from ipywidgets import IntSlider, FloatSlider, interact
        self._interaction_state = json.loads(self._interaction_file.read_text())

        def show_anim(**kwargs):
            self._interaction_state = kwargs
            self._interaction_file.write_text(json.dumps(kwargs))
            self.preview(kwargs["i"])

        vars = dict(i=IntSlider(
            min=0,
            max=self.duration-1,
            continuous_update=False,
            value=self._interaction_state.get("i", 0),
            description="f.i"))
    
        for v, val in self.vars.items():
            vars[v] = FloatSlider(min=0,
                max=1, step=1e-2,
                continuous_update=False,
                value=self._interaction_state.get(v, val))
        
        interact(show_anim, **vars)
    
    def iv(self, k):
        return self._interaction_state.get(k, 0)
    
    def preview(self, *frames):
        if len(frames) == 0:
            frames = [0]
        elif len(frames) == 1:
            if frames[0] == "*":
                frames = list(range(self.start, self.end))
        
        for frame in frames:
            show_frame(self, frame, scale=self.preview_scale)
        return self
    
    def render(self, scale=1):
        render_animation(self, show=[], scale=scale)
        return self
    
    def show(self, loops=1, verbose=False, download=False, scale=0.5):
        if self.fmt == "svg":
            show_animation(self, start=False)
        else:
            show_video(self, loops=loops, verbose=verbose, download=download, scale=scale)
        return self
    
    def zip(self, download=False):
        zipfile = f"{str(self.output_folder)}.zip"
        run(["zip", "-j", "-r", zipfile, str(self.output_folder)])
        print("> zipped:", zipfile)
        if download:
            try:
                from google.colab import files
                files.download(zipfile)
            except ImportError:
                print("download= arg is for colab")
                pass
        return self


class notebook_aframe(aframe):
    def __init__(self,
        rect=(540, 540),
        **kwargs
        ):
        self._preview = [0]

        super().__init__(rect,
            timeline=Timeline(1),
            interactive=False,
            **kwargs)


import unittest

import coldtype.pens.misc
from coldtype.pens.reportlabpen import ReportLabPen
from coldtype.test import StSt, co, Rect, DP, DPS, hsl


class TestReportLabPen(unittest.TestCase):
    def test_basic_drawing(self):
        coldtype.pens.misc.USE_SKIA_PATHOPS = False
        r = Rect("letter")
        pens = DPS([
            (DP(r.inset(20))
                .align(r)
                .f(0.9)),
            (DP(r.inset(40))
                .align(r)
                .f(None)
                .s(0).sw(5)),
            (StSt("COLD", co, 150, wdth=1)
                .align(r)
                .f(hsl(0.3))
                .s(hsl(0.9))
                .sw(5)
                .pen()
                .q2c()
                .removeOverlap()
                )])
        
        self.assertEqual(len(pens[-1].value), 117)
        ReportLabPen.PDF(pens, r,
            "test/ignorables/reportlab_test.pdf")
        coldtype.pens.misc.USE_SKIA_PATHOPS = True

if __name__ == "__main__":
    unittest.main()


import unittest
from pathlib import Path
from coldtype.renderable import renderable, Action
from coldtype.renderer.reader import SourceReader


test_src = """
from coldtype import *

@renderable()
def test_src_function(r):
    return (StSt("CDELOPTY",
        "assets/ColdtypeObviously-VF.ttf")
        -.align(r.inset(50))
        .f(hsl(0.5, 0.7, 0.9)))

@animation()
def test_src_animation(f):
    return (StSt("CDELOPT",
        "assets/ColdtypeObviously-VF.ttf",
        wdth=f.e("linear", 1))
        .align(f.a.r))
"""


class TestSyntaxMods(unittest.TestCase):
    def setUp(self) -> None:
        self.sr = SourceReader(None, test_src)
        self.sr2 = SourceReader(runner="special")
        return super().setUp()

    def tearDown(self) -> None:
        self.sr.unlink()
        self.sr2.unlink()
        return super().tearDown()
    
    def test_empty_reader(self):
        self.assertEqual(self.sr2.filepath, None)
        self.assertEqual(self.sr2.codepath, None)
        self.sr2.reset_filepath("test/source_file.py")
        self.assertEqual(self.sr2.filepath.stem, "source_file")
        self.assertEqual(len(self.sr2.renderables()), 1)
        self.assertEqual(self.sr2.program["__RUNNER__"], "special")
    
    def test_frame_read(self):
        result = self.sr.frame_results(1)
        self.assertEqual(len(result), 2)
        self.assertEqual(len(result[0]), 2)
        self.assertEqual(len(result[0][1]), 8)

    def test_syntax_mods(self):
        sr = self.sr
        mod_src = sr.codepath.read_text()

        self.assertIn(".align(r.inset(50))", test_src)
        self.assertNotIn(".align(r.inset(50))", mod_src)
        self.assertIn(".noop()", mod_src)

        self.assertEqual(sr.program["__RUNNER__"], "default")

        renderables = sr.renderables(output_folder_override="test/renders")
        self.assertEqual(len(renderables), 2)
        self.assertEqual(renderables[0].codepath, sr.codepath)

        r1:renderable = renderables[0]
        r1_passes = r1.passes(None, None, [])
        r1p1_result = r1.run_normal(r1_passes[0])
        self.assertEqual(r1p1_result[3].glyphName, "L")

        r2:renderable = renderables[1]
        r2_passes = r2.passes(Action.RenderAll, None, [])
        for idx, rp in enumerate(r2_passes):
            rel = rp.output_path.relative_to(Path.cwd() / "test/renders")
            self.assertEqual(
                "test_src_animation_{:04}.png".format(idx),
                str(rel))
        
        # verify that the width of the 'C' is increasing over the first half of the animation
        r2p2_result = r2.run_normal(r2_passes[1])
        r2p3_result = r2.run_normal(r2_passes[2])
        r2p4_result = r2.run_normal(r2_passes[3])
        self.assertLess(
            r2p2_result[0].ambit().w,
            r2p3_result[0].ambit().w)
        self.assertLess(
            r2p3_result[0].ambit().w,
            r2p4_result[0].ambit().w)

        renderables = sr.renderables(viewer_solos=[1])
        self.assertEqual(len(renderables), 1)
        self.assertEqual(renderables[0].name, "test_src_animation")
    
        # indexes over len wrap-around
        renderables = sr.renderables(viewer_solos=[2])
        self.assertEqual(len(renderables), 1)
        self.assertEqual(renderables[0].name, "test_src_function")

        renderables = sr.renderables(function_filters=[r".*_function"])
        self.assertEqual(len(renderables), 1)
        self.assertEqual(renderables[0].name, "test_src_function")

        renderables = sr.renderables(function_filters=[r".*_animation"])
        self.assertEqual(len(renderables), 1)
        self.assertNotEqual(renderables[0].name, "test_src_function")

        renderables = sr.renderables(class_filters=[r".*asdf$"])
        self.assertEqual(len(renderables), 0)

        renderables = sr.renderables(class_filters=[r".*ation$"])
        self.assertEqual(len(renderables), 1)
        self.assertEqual(renderables[0].__class__.__name__, "animation")

        # when the pattern matches nothing, all renderables returned
        renderables = sr.renderables(function_filters=[r".*_should_be_nothing"])
        self.assertEqual(len(renderables), 2)
    
        sr.reload(test_src.replace("-.align(", ".align("))
        mod_src = sr.codepath.read_text()
        self.assertNotIn(".noop()", mod_src)
        self.assertIn(".align(r.inset(50))", mod_src)

        prev_codepath = sr.codepath
        prev_filepath = sr.filepath
        test_source_file = Path("test/source_file.py")
        sr.reset_filepath("test/source_file.py")

        self.assertEqual(sr.filepath.absolute(), test_source_file.absolute())
        self.assertNotEqual(sr.codepath, prev_codepath)

        self.assertTrue(not prev_filepath.exists())
        self.assertTrue(not prev_codepath.exists())
        self.assertTrue(sr.codepath.exists())

        sr.unlink()

        self.assertTrue(not sr.codepath.exists())
        self.assertTrue(test_source_file.exists())

        sr.reset_filepath("test/source_file")
        self.assertEqual(sr.filepath.suffix, ".py")
        self.assertEqual(sr.filepath, test_source_file.absolute())

if __name__ == "__main__":
    unittest.main()


import re, os
from pathlib import Path
from runpy import run_path
from functools import partial
from tempfile import NamedTemporaryFile

from coldtype.renderable import renderable
from coldtype.renderable.animation import animation

from coldtype.renderer.utils import Watchable
from coldtype.renderer.config import ColdtypeConfig
from coldtype.helpers import sibling
from coldtype.text.reader import ALL_FONT_DIRS

try:
    from docutils.core import publish_doctree
except ImportError:
    publish_doctree = None


def apply_syntax_mods(source_code, renderer=None):
    codepath_offset = 0

    def inline_arg(p):
        path = Path(inline.strip()).expanduser().absolute()
        if renderer:
            if path not in renderer.watchee_paths():
                renderer.watchees.append([Watchable.Source, path, None])
        src = path.read_text()
        codepath_offset = len(src.split("\n"))
        return src
    
    def inline_other(x):
        cwd = Path.cwd()
        path = Path(cwd / (x.group(1).replace(".", "/")+".py"))
        if renderer:
            if path not in renderer.watchee_paths():
                renderer.watchees.append([Watchable.Source, path, None])
        src = path.read_text()
        codepath_offset = len(src.split("\n"))
        return src

    if renderer and renderer.source_reader.config.inline_files:
        for inline in renderer.source_reader.config.inline_files:
            source_code = inline_arg(inline) + "\n" + source_code

    source_code = re.sub(r"from ([^\s]+) import \* \#INLINE", inline_other, source_code)
    source_code = re.sub(r"\-\.[A-Za-z_ƒ]+([A-Za-z_0-9]+)?\(", ".nerp(", source_code)
    source_code = re.sub(r"([\s]+)Ƨ\(", r"\1nerp(", source_code)
    #source_code = re.sub(r"λ[\s]{0,3}\.", "lambda p: p.", source_code)
    #source_code = re.sub(r"λ\s?([/\.\@]{1,2})", r"lambda xxx: xxx\1", source_code)
    #source_code = re.sub(r"λ\.", "lambda x: x.", source_code)
    #source_code = re.sub(r"λ", "lambda ", source_code)
    #source_code = re.sub(r"ßDPS\(([^\)]+)\)", r"(ß:=DPS(\1))", source_code)

    while "nerp(" in source_code:
        start = source_code.find("nerp(")
        end = -1
        i = 5
        depth = 1
        c = source_code[start+i]
        while depth > 0 and c:
            #print(c, depth)
            if c == "(":
                depth += 1
            elif c== ")":
                depth -= 1
            i += 1
            c = source_code[start+i]
        end = start+i
        source_code = source_code[:start] + "noop()" + source_code[end:]
        #print(start, end)
    
    if renderer:
        renderer._codepath_offset = codepath_offset
    
    return source_code, codepath_offset


def read_source_to_tempfile(filepath:Path,
    codepath:Path=None,
    renderer=None
    ):
    data_out = {}

    if filepath.suffix == ".rst":
        if not publish_doctree:
            raise Exception("pip install docutils")
        doctree = publish_doctree(filepath.read_text())
        def is_code_block(node):
            if node.tagname == "literal_block":
                classes = node.attributes["classes"]
                # ok the "ruby" here is an ugly hack but it's so I can hide certain code
                # from a printed rst
                if "code" in classes and ("python" in classes or "ruby" in classes):
                    return True
            return False
        code_blocks = doctree.traverse(condition=is_code_block)
        source_code = [block.astext() for block in code_blocks]
        if codepath and codepath.exists():
            codepath.unlink()
        with NamedTemporaryFile("w", prefix="coldtype_rst_src", suffix=".py", delete=False) as tf:
            tf.write("\n".join(source_code))
            codepath = Path(tf.name)
    
    elif filepath.suffix == ".md":
        from lxml.html import fragment_fromstring, tostring
        try:
            import markdown
        except ImportError:
            raise Exception("pip install markdown")
        try:
            import frontmatter
        except ImportError:
            frontmatter = None
            print("> pip install python-frontmatter")
        md = markdown.markdown(filepath.read_text(),
            extensions=["fenced_code"])
        fm = frontmatter.loads(filepath.read_text())
        data_out["frontmatter"] = fm
        frag = fragment_fromstring(md, create_parent=True)
        blocks = []
        for python in frag.findall("./pre/code[@class='python']"):
            blocks.append(python.text)
        source_code = "\n".join(blocks)
        if codepath and codepath.exists():
            codepath.unlink()
        with NamedTemporaryFile("w", prefix="coldtype_md_src", suffix=".py", delete=False) as tf:
            mod_src, _ = apply_syntax_mods(source_code, renderer)
            tf.write(mod_src)
            codepath = Path(tf.name)
    
    elif filepath.suffix == ".py":
        source_code, _ = apply_syntax_mods(filepath.read_text(), renderer)
        if codepath and codepath.exists():
            codepath.unlink()
        with NamedTemporaryFile("w", prefix=f"coldtype__{filepath.stem}_", suffix=".py", delete=False) as tf:
            tf.write(source_code)
            codepath = Path(tf.name)
    else:
        raise Exception("No code found in file!")
    
    return codepath, data_out


def run_source(filepath, codepath, **kwargs):
    return run_path(str(codepath), init_globals={
        "__COLDTYPE__": True,
        "__FILE__": filepath,
        "__sibling__": partial(sibling, filepath),
        **kwargs})


def renderable_to_output_folder(filepath, renderable, override=None):
    if override:
        return Path(override).expanduser().resolve()
    elif renderable.dst:
        return renderable.dst / (renderable.custom_folder or renderable.folder(filepath))
    else:
        return (filepath.parent if filepath else Path(os.getcwd())) / "renders" / (renderable.custom_folder or renderable.folder(filepath))


def find_renderables(
    filepath:Path,
    codepath:Path,
    program:dict,
    viewer_solos=[],
    function_filters=[],
    class_filters=[],
    output_folder_override=None,
    ):
    all_rs = []
    filtered_rs = []
    
    for k, v in program.items():
        if isinstance(v, renderable) and not v.hidden:
            if v not in all_rs:
                all_rs.append(v)
        elif k == "RENDERABLES":
            for r in v:
                all_rs.append(r)
    
    #all_rs = sorted(all_rs, key=lambda r: r.layer)
    all_rs = sorted(all_rs, key=lambda r: r.sort)

    for r in all_rs:
        r.filepath = filepath
        r.codepath = codepath
        r.output_folder = renderable_to_output_folder(
            filepath, r, override=output_folder_override)
        r.post_read()

    if any([r.solo for r in all_rs]):
        filtered_rs = [r for r in all_rs if r.solo]
    else:
        filtered_rs = all_rs
        
    if function_filters:
        function_patterns = function_filters
        matches = []
        for r in filtered_rs:
            for fp in function_patterns:
                try:
                    if re.search(fp, r.name) and r not in matches:
                        matches.append(r)
                except re.error as e:
                    print("ff regex compilation error", e)
        if len(matches) > 0:
            filtered_rs = matches
    
    if class_filters:
        matches = []
        for r in filtered_rs:
            for cf in class_filters:
                try:
                    if re.match(cf, r.__class__.__name__) and r not in matches:
                        matches.append(r)
                except re.error as e:
                    print("cf regex compilation error", e)
        filtered_rs = matches
    
    if len(viewer_solos) > 0:
        viewer_solos = [vs%len(filtered_rs) for vs in viewer_solos]

        solos = []
        for i, r in enumerate(filtered_rs):
            if i in viewer_solos:
                solos.append(r)
        filtered_rs = solos
    
    return filtered_rs


class SourceReader():
    def __init__(self,
        filepath:Path=None,
        code:str=None,
        renderer=None,
        runner:str="default",
        cli_args=None,
        ):
        self.filepath = None
        self.codepath = None
        self.data_out = None
        self.dirpath = None
        self.should_unlink = False
        self.program = None
        self.renderer = renderer
        self.runner = runner

        self.config = None
        self.read_configs(cli_args, filepath)

        if filepath or code:
            self.reset_filepath(filepath, code)
    
    def read_configs(self, args, filepath):
        embedded = Path(__file__).parent / ".coldtype.py"
        proj = Path(".coldtype.py")
        user = Path("~/.coldtype.py").expanduser()
        files = [embedded, user, proj]
        if args and hasattr(args, "config") and args.config:
            if args.config == "0":
                files = []
            else:
                if args.config == ".":
                    args.config = args.file
                files.append(Path(args.config).expanduser())

        if filepath:
            files.append(Path(filepath).expanduser())
        
        if args and args.file:
            fp = Path(args.file).expanduser()
            if fp.exists() and not fp.is_dir():
                files.append(fp)

        py_config = {}
        for p in files:
            if p.exists():
                try:
                    py_config = run_path(str(p), init_globals={
                        "__FILE__": p,
                        "__sibling__": partial(sibling, p),
                    })
                    #self.midi_mapping = py_config.get("MIDI", self.midi_mapping)
                    #self.hotkey_mapping = py_config.get("HOTKEYS", self.hotkey_mapping)
                    self.config = ColdtypeConfig(py_config, self.config if self.config else None, args)

                    for f in self.config.font_dirs:
                        ALL_FONT_DIRS.insert(0, f)
                except Exception as e:
                    print("Failed to load config", p)
                    print("Exception:", e)
        
        if len(files) == 0 or not self.config:
            self.config = ColdtypeConfig({}, None, args)
        
        self.config.args = args
        #print(self.config.values())
    
    def find_sources(self, dirpath):
        sources = list(dirpath.glob("*.py"))
        sources.extend(list(dirpath.glob("*.md")))

        valid_sources = []
        for p in sources:
            if not p.name.startswith("_"):
                valid_sources.append(p)
            valid_sources = sorted(valid_sources, key=lambda p: p.stem)
        
        return valid_sources
    
    def normalize_filepath(self, filepath:Path, dirindex=0):
        if isinstance(filepath, str):
            filepath = Path(filepath)
        
        filepath = filepath.expanduser().resolve()

        if filepath.is_dir():
            self.dirpath = filepath
            #filepath = sorted(list(filepath.glob("*.py")), key=lambda p: p.stem)[dirindex]
            filepath = self.find_sources(self.dirpath)[dirindex]
        else:
            self.dirpath = filepath.parent

        if not filepath.exists():
            with_py = (filepath.parent / (filepath.stem + ".py"))
            if with_py.exists():
                filepath = with_py
        if filepath.suffix not in [".md", ".rst", ".py"]:
            raise Exception("Coldtype can only read .py, .md, and .rst files")
        
        return filepath
    
    def reset_filepath(self, filepath:Path, code:str=None, reload:bool=True, dirdirection=0):
        self.unlink()
        self.should_unlink = False

        dirindex = 0
        if self.dirpath and dirdirection != 0:
            # find index of existing filepath, increment by dirdirection?
            pys = self.find_sources(self.dirpath)
            #pys = sorted(list(self.dirpath.glob("*.py")), key=lambda p: p.stem)
            curr = pys.index(self.filepath)
            adj = (curr + dirdirection) % len(pys)
            print(curr, adj)
            filepath = pys[adj]
        
        if filepath:
            self.filepath = self.normalize_filepath(filepath, dirindex)
            if not self.filepath.exists():
                raise Exception("Source file does not exist")
        else:
            if code:
                self.write_code_to_tmpfile(code)
            else:
                raise Exception("Must provide filepath or code")
            
        if reload:
            self.reload()
        
        return self.filepath
    
    def reload(self, code:str=None):
        if code:
            self.write_code_to_tmpfile(code)
        
        self.codepath, self.data_out = read_source_to_tempfile(self.filepath, self.codepath, renderer=self.renderer)
        self.program = run_source(
            self.filepath,
            self.codepath,
            __RUNNER__=self.runner)
    
    def write_code_to_tmpfile(self, code):
        if self.filepath:
            self.filepath.unlink()

        with NamedTemporaryFile("w", prefix="coldtype_", suffix=".py", delete=False) as tf:
            tf.write(code)
        
        self.filepath = Path(tf.name)
        self.should_unlink = True
    
    def renderables(self,
        viewer_solos=[],
        function_filters=[],
        class_filters=[],
        output_folder_override=None,
        ):
        if not function_filters and self.config.function_filters:
            function_filters = self.config.function_filters
        return find_renderables(
            self.filepath,
            self.codepath,
            self.program,
            viewer_solos,
            function_filters,
            class_filters,
            output_folder_override)
    
    def frame_results(self, frame, class_filters=[]):
        rs = self.renderables(class_filters=class_filters)
        res = []
        for r in rs:
            ps = r.passes(None, None, indices=[frame])
            for p in ps:
                res.append([r, r.run_normal(p)])
        return res
    
    def unlink(self):
        if self.should_unlink and self.filepath:
            if self.filepath.exists():
                self.filepath.unlink()
        if self.codepath:
            if self.codepath.exists():
                self.codepath.unlink()
        return self


import unittest
from pathlib import Path
from coldtype.grid import Grid
from coldtype.geometry import *
from coldtype.color import hsl
from coldtype.text.composer import StSt, Glyphwise, Style, Font
from coldtype.pens.draftingpens import DraftingPens
from coldtype.pens.svgpen import SVGPen

tf = Path(__file__).parent

class TestGlyphwise(unittest.TestCase):
    def _test_glyph_kerning(self, font_path, kern, pj=False):
        txt = "AVAMANAV"
        #txt = "AVAV"
        ss = StSt(txt, font_path, 100, wdth=0, kern=kern)
        gw = Glyphwise(txt, lambda g: Style(font_path, 100, wdth=0, kern=kern, ro=1))
        gwo = Glyphwise(txt, lambda g: Style(font_path, 100, wdth=0, kern=(not kern), ro=1))

        self.assertEqual(len(ss), len(txt))
        self.assertEqual(ss[0].glyphName, "A")
        self.assertEqual(ss[-1].glyphName, "V")
        self.assertEqual(len(gw), len(txt))
        self.assertEqual(ss[0].glyphName, "A")
        self.assertEqual(ss[-1].glyphName, "V")

        if pj:
            r = Rect(1500, 500)
            DraftingPens([
                ss,
                gw.f(None).s(0).sw(5),
                gwo.copy().f(None).s(hsl(0.9)).sw(5),
            ]).translate(20, 20).scale(5, point=Point(0, 0)).picklejar(r, name=f"gw_kern_{kern}")

        self.assertEqual(ss.ambit(), gw.ambit())
        self.assertNotEqual(ss.ambit(), gwo.ambit())

        fp = Path(font_path)
        op = (tf / f"ignorables/__{fp.name}.svg")
        op.parent.mkdir(exist_ok=True)
        op.write_text(SVGPen.Composite(DraftingPens([ss, gw.translate(0, 10)]), ss.ambit(), viewBox=True))

        return ss, gw

    def test_format_equality(self):
        fnt = "~/Type/fonts/fonts/OhnoFatfaceVariable.ttf"
        self._test_glyph_kerning(fnt, False, pj=False)
        self._test_glyph_kerning(fnt, True, pj=True)
    
    def test_ligature(self):
        clarette = Font.Cacheable("~/Type/fonts/fonts/_wdths/ClaretteGX.ttf")
        r = Rect(1080, 300)
        gl = (Glyphwise(["fi", "j", "o", "ff"],
            lambda g: Style(clarette, 200, wdth=g.i/3))
            .align(r))
        
        self.assertEqual(len(gl), 4)
        self.assertEqual(gl[0].glyphName, "f_i")
        self.assertEqual(gl[-1].glyphName, "f_f")
        gl.picklejar(r)
    
    # def test_grouped(self):
    #     fnt = "~/Type/fonts/fonts/OhnoFatfaceVariable.ttf"
    #     r = Rect(1000, 200)
    #     gw = (Glyphwise(["AB", "CD"], lambda g:
    #         Style(fnt, 250, wdth=i, opsz=i))
    #         .align(r))
        
    #     gw.picklejar(r)

    def test_variable_args(self):
        fnt = Font.Find("OhnoFatfaceV")
        r = Rect(1080, 300)
        
        (Glyphwise("FATFACE 1 ARG", lambda g:
            Style(fnt, 200, wdth=g.e))
            .align(r)
            .picklejar(r))
        
        es = []
        def print_e(i, g):
            es.append(g.e)
            return Style(fnt, 200, opsz=g.e, wdth=1-g.e)
        
        (Glyphwise("FATFACE 2 ARG", print_e)
            .align(r)
            .picklejar(r))
        
        self.assertEqual(es[0], 0)
        self.assertEqual(es[-1], 1)
    
    def test_newline(self):
        fnt = Font.Find("OhnoFatfaceV")
        r = Rect(1080, 300)
        
        gs = (Glyphwise("FATFACE\nFACEFAT", lambda g:
            Style(fnt, g.i*10+50, wdth=g.e))
            .align(r)
            .picklejar(r))
        
        self.assertEqual(gs[0][0].ambit().xy(),
            [305.21250000000003, 172.05])
        self.assertEqual(gs[1][-1].ambit().xy(),
            [656.3775, 58.650000000000006])
        self.assertEqual(gs[0][-1].glyphName, "E")
        self.assertEqual(gs[1][-1].glyphName, "T")
    
    def test_kp(self):
        fnt = Font.Find("OhnoFatfaceV")
        r = Rect(1080, 300)

        gs_no_kp = (Glyphwise("FATFACE", lambda g:
            Style(fnt, 250, wdth=1-g.e, ro=1))
            .align(r)
            .fssw(-1, 0, 1)
            .picklejar(r))
        
        gs_kp = (Glyphwise("FATFACE", lambda g:
            Style(fnt, 250, wdth=1-g.e, kp={"A/T":-250}, ro=1))
            .align(r)
            .fssw(-1, 0, 1)
            .picklejar(r))
        
        kp_sw = gs_kp[2].ambit().psw
        no_kp_sw = gs_no_kp[2].ambit().psw
    
        self.assertNotEqual(kp_sw, no_kp_sw)
        self.assertEqual(kp_sw.y, no_kp_sw.y)
    
    def test_tu(self):
        fnt = Font.Find("OhnoFatfaceV")
        r = Rect(1080, 300)

        gs_no_tu = (Glyphwise("FATFACE", lambda g:
            Style(fnt, 250, wdth=1-g.e, ro=1))
            .align(r)
            .fssw(-1, 0, 1)
            .picklejar(r))
        
        gs_tu = (Glyphwise("FATFACE", lambda g:
            Style(fnt, 250, wdth=1-g.e, tu=-50, ro=1))
            .align(r)
            .fssw(-1, 0, 1)
            .picklejar(r))
    
        self.assertLess(gs_tu.ambit().w, gs_no_tu.ambit().w)

if __name__ == "__main__":
    unittest.main()


import threading, json

from pathlib import Path
from http.server import SimpleHTTPRequestHandler, HTTPServer

from coldtype.renderer.winman.passthrough import WinmanPassthrough
from coldtype.renderer.config import ConfigOption, ColdtypeConfig

from coldtype.pens.svgpen import SVGPen
from coldtype.pens.jsonpen import JSONPen

WEBSOCKET_PORT = None

class WebViewerHandler(SimpleHTTPRequestHandler):
    def _set_headers(self):
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()

    def do_GET(self):
        self._set_headers()
        self.wfile.write(
            (Path(__file__).parent.parent.parent / "webserver/webviewer.html")
                .read_text()
                .replace("localhost:8007", f"localhost:{WEBSOCKET_PORT}")
                .encode("utf8"))

    def do_HEAD(self):
        self._set_headers()
    
    def log_message(self, format, *args):
        pass

class WinmanWebview(WinmanPassthrough):
    def __init__(self, config:ColdtypeConfig, renderer):
        self.config = config
        self.renderer = renderer

        global WEBSOCKET_PORT
        WEBSOCKET_PORT = self.config.websocket_port

        wv_port = self.config.webviewer_port
        if wv_port != 0:
            print("WEBVIEWER>", f"localhost:{wv_port}")

            def start_server(port):
                httpd = HTTPServer(('', port), WebViewerHandler)
                httpd.serve_forever()

            daemon = threading.Thread(name='daemon_server',
                target=start_server, args=(wv_port,))
            daemon.setDaemon(True)
            daemon.start()
    
    def turn_over(self):
        renders = []
        try:
            title = self.renderer.watchees[0][1].name
        except:
            title = "coldtype"

        for idx, (render, result, rp) in enumerate(self.renderer.previews_waiting_to_paint):
            if self.renderer.args.format == "canvas": # TODO config?
                renders.append(dict(
                    fmt="canvas",
                    jsonpen=JSONPen.Composite(result, render.rect),
                    rect=[*render.rect],
                    bg=[*render.bg]))
            else:
                renders.append(dict(
                    fmt="svg", svg=SVGPen.Composite(
                        result,
                        render.rect,
                        viewBox=render.viewBox),
                    rect=[*render.rect],
                    bg=[*render.bg]))
    
        if renders:
            for _, client in self.renderer.server.connections.items():
                if hasattr(client, "webviewer") and client.webviewer:
                    client.sendMessage(json.dumps({
                        "renders":renders,
                        "title":title
                    }))
        
        return []




class WinmanPassthrough():
    def __init__(self):
        self.context = None

    def set_title(self, text):
        #print("TITLE", text)
        pass
    
    def terminate(self):
        pass

    def reset(self):
        pass

    def preview_scale(self):
        return 1
    
    def turn_over(self):
        print("turning")
    
    def should_close(self):
        return False


import unittest
from coldtype import Rect, Font, Style, DP, Path, hsl
from coldtype.text.richtext import RichText
from coldtype.remote import show_picklejar

f1 = Font("assets/ColdtypeObviously-VF.ttf")
f2 = Font("assets/MutatorSans.ttf")

class TestRichText(unittest.TestCase):
    def test_preserve_space(self):
        r = Rect(1200, 300)
        rt = RichText(r, "HELLO[i] COLDTYPE", dict(
            i=Style(f2, 200, wdth=0, wght=1),
            default=Style(f1, 200, wdth=0))).align(r)

        self.assertEqual(rt[0][0].data.get("txt"), "COLDTYPE")
        self.assertEqual(rt[0][1].data.get("txt"), "HELLO ")

        self.assertEqual(rt[0][1][0].glyphName, "space")
        self.assertEqual(rt[0][1][-1].glyphName, "H")

        self.assertEqual(rt[0][0][0].glyphName, "E")
        self.assertEqual(rt[0][0][-1].glyphName, "C")

        space_width = rt[0][1][0].ambit(th=0).w
        self.assertEqual(space_width, 50)

        self.assertGreater(
            rt[0][1].ambit(th=0).w - space_width,
            rt[0][1].ambit(th=1).w)
        
        rt.picklejar(r)
    
    def test_ligature(self):
        clarette = Font.Cacheable("~/Type/fonts/fonts/_wdths/ClaretteGX.ttf")

        txt = "fi¬joff≤asdf≥"
        r = Rect(1080, 300)
        
        gl = (RichText(r, txt,
            dict(
                default=Style(clarette, 200),
                asdf=Style(clarette, 200, wdth=1)),
            tag_delimiters=["≤", "≥"],
            visible_boundaries=["¶"],
            invisible_boundaries=["¬"])
            .align(r))
        
        self.assertEqual(gl[0][0].data["style_names"][0], "asdf")
        self.assertEqual(len(gl[0][1].data["style_names"]), 0)
        self.assertEqual(gl[0][0][0].glyphName, "f_f")
        self.assertEqual(gl[0][1][0].glyphName, "f_i")
        
        gl.picklejar(r)

if __name__ == "__main__":
    unittest.main()


import unittest
from coldtype.color import *

class TestColor(unittest.TestCase):
    def test_interp(self):
        _hsl = hsl(0.1).hsl_interp(0, hsl(0.9))
        self.assertAlmostEqual(_hsl.hp, 0.1)

        _hsl = hsl(0.1).hsl_interp(1, hsl(0.9))
        self.assertAlmostEqual(_hsl.hp, 0.9)

        _hsl = hsl(0.2).hsl_interp(0.5, hsl(0.6))
        self.assertAlmostEqual(_hsl.hp, 0.4)

if __name__ == "__main__":
    unittest.main()



try:
    import rtmidi
except ImportError:
    rtmidi = None
    pass

class MIDIWatcher():
    def __init__(self, config, state, on_shortcut):
        self.config = config
        self.state = state
        self.on_shortcut = on_shortcut
        self.failed = False
        self.devices = []

        try:
            midiin = rtmidi.RtMidiIn()
            lookup = {}
            for p in range(midiin.getPortCount()):
                lookup[midiin.getPortName(p)] = p

            for device, mapping in self.config.midi.items():
                if device in lookup:
                    mapping["port"] = lookup[device]
                    mi = rtmidi.RtMidiIn()
                    mi.openPort(lookup[device])
                    self.devices.append([device, mi])
                else:
                    if self.config.midi_info:
                        print(f">>> no midi port found with that name ({device}) <<<")
            
            if self.config.midi_info:
                print("\nMIDI DEVICES:::")
                midiin = rtmidi.RtMidiIn()
                ports = range(midiin.getPortCount())
                for p in ports:
                    print(">", f"\"{midiin.getPortName(p)}\"")
                print("\n")

        except Exception as e:
            self.failed = True
            print("MIDI SETUP EXCEPTION >", e)
    
    def monitor(self, playing):
        if self.failed:
            return

        controllers = {}
        for device, mi in self.devices:
            msg = mi.getMessage(0)
            while msg:
                if self.config.midi_info:
                    print(device, msg)
                if msg.isNoteOn(): # Maybe not forever?
                    nn = msg.getNoteNumber()
                    shortcut = self.config.midi[device]["note_on"].get(nn)
                    self.on_shortcut(shortcut, nn)
                if msg.isController():
                    #print(">>>", msg)
                    cn = msg.getControllerNumber()
                    cv = msg.getControllerValue()
                    cc = msg.getChannel()
                    shortcut = self.config.midi[device].get("controller", {}).get(cn)
                    if shortcut:
                        if cv in shortcut:
                            print("shortcut!", shortcut, cv)
                            self.on_shortcut(shortcut.get(cv), cn)
                    else:
                        controllers["_".join([device, str(cc), str(cn)])] = cv
                msg = mi.getMessage(0)
        
        if len(controllers) > 0:
            nested = {}
            for k, v in controllers.items():
                device, channel, number = k.split("_")
                if not nested.get(device):
                    nested[device] = {}
                if not nested[device].get(channel):
                    nested[device][channel] = {}
                nested[device][channel][str(number)] = v
            
            for device, channels in nested.items():
                if not self.state.controller_values.get(device):
                    self.state.controller_values[device] = {}
                for channel, numbers in channels.items():
                    self.state.controller_values[device][channel] = {**self.state.controller_values.get(device, {}).get(channel, {}), **numbers}

            if not playing:
                return True


from pathlib import Path

from coldtype.renderer.winman.passthrough import WinmanPassthrough
from coldtype.blender.render import blender_launch_livecode

class WinmanBlender(WinmanPassthrough):
    def __init__(self):
        pass
        self.subp = None
    
    def launch(self, blend_file):
        if self.subp:
            self.subp.kill()
        self.subp = blender_launch_livecode(blend_file)
    
    def write_command(self, cmd, arg):
        try:
            cb = Path("~/.coldtype/blender.txt").expanduser()
            if cb.exists():
                cb.unlink()
            cb.write_text(f"{cmd},{str(arg)}")
        except FileNotFoundError:
            pass
    
    def reload(self, filepath):
        self.write_command("import", filepath)
    
    def toggle_playback(self, toggle):
        self.write_command("play_preview", toggle)
    
    def frame_offset(self, offset):
        self.write_command("frame_offset", offset)

    def terminate(self):
        if self.subp:
            self.subp.kill()


import unittest
from random import Random
from coldtype.geometry import Rect, Point
from coldtype.pens.draftingpen import DraftingPen
from coldtype.pens.draftingpens import DraftingPens
from coldtype.color import hsl, rgb
from coldtype.pens.drawablepen import DrawablePenMixin
from coldtype.renderer.reader import SourceReader
from coldtype.text.composer import StSt, Font

co = Font.Cacheable("assets/ColdtypeObviously-VF.ttf")
mutator = Font.Cacheable("assets/MutatorSans.ttf")

class TestDraftingPens(unittest.TestCase):
    def test_gs(self):
        r = Rect(0, 0, 100, 100)
        dps = DraftingPens()
        dp = (DraftingPen()
            .define(r=r)
            .gs("$r↖ $r↗ $r↙|↘|65 ɜ"))
        self.assertEqual(len(dp.value), 4)
        self.assertEqual(dp.value[-2][-1][0], Point(100, 35))
        self.assertEqual(dp.value[-1][0], "endPath")
        self.assertEqual(dp.unended(), False)
        dps.append(DraftingPens([dp]))
        self.assertEqual(len(dps.tree().splitlines()), 3)
        self.assertEqual(dps.tree().splitlines()[-1],
            " | | DraftingPen<4mvs:end/>")
        
    def test_gs_arrowcluster(self):
        r = Rect(100, 100)
        dp = (DraftingPen()
            .define(r=r)
            .gs("$r↖↗↘"))
        
        self.assertEqual(len(dp.value), 4)
        self.assertEqual(dp.value[0][-1][0], Point(0, 100))
        self.assertEqual(dp.value[1][-1][0], Point(100, 100))
        self.assertEqual(dp.value[2][-1][0], Point(100, 0))
    
    def test_gs_relative_moves(self):
        r = Rect(100, 100)
        dp = (DraftingPen()
            .define(r=r)
            .gs("$r↖ ¬OX50OY-50 §OY-50"))
        
        self.assertEqual(len(dp.value), 4)
        self.assertEqual(dp.value[0][-1][0], Point(0, 100))
        self.assertEqual(dp.value[1][-1][0], Point(50, 50))
        self.assertEqual(dp.value[2][-1][0], Point(0, 50))
    
    def test_gss(self):
        """
        A rect passed to gs and gss should create the same value on the pen
        """
        dp1 = (DraftingPen()
            .define(r=Rect(100, 100))
            .gss("$r"))
        
        dp2 = (DraftingPen()
            .define(r=Rect(100, 100))
            .gs("$r"))
        
        self.assertEqual(dp1.value, dp2.value)
        
    def test_reverse(self):
        dp = (DraftingPen()
            .define(r=Rect(100, 100))
            .gs("$r↖ $r↗ $r↘ ɜ"))
        p1 = dp.value[0][-1]
        p2 = dp.reverse().value[-2][-1]
        self.assertEqual(p1, p2)
    
    def test_transforms(self):
        dp = (DraftingPen(Rect(100, 100))
            .frame(Rect(100, 100))
            .align(Rect(200, 200)))
        
        self.assertEqual(dp.frame().mxx, 150)
        self.assertEqual(dp.value[-2][-1][-1][0], 50)

        self.assertEqual(
            dp.copy().rotate(45).round().value,
            dp.copy().rotate(360+45).round().value)
        
        self.assertEqual(dp.copy().scale(2).ambit().w, 200)

    def test_pens_ambit(self):
        dps = (DraftingPens([
                DraftingPen(Rect(50, 50)),
                DraftingPen(Rect(100, 100, 100, 100))])
                #.print(lambda x: x.tree())
                )
        ram = dps.ambit()
        self.assertEqual(ram, Rect(0, 0, 200, 200))

        moves = []
        dps.walk(lambda p, pos, _: moves.append([p, pos]))
        self.assertEqual(moves[0][0], dps)
        self.assertEqual(moves[0][1], -1)
        self.assertEqual(moves[1][1], 0)
    
    def test_remove_blanks(self):
        dps = (DraftingPens([
            DraftingPen(Rect(50, 50)),
            DraftingPen()
        ]))
        self.assertEqual(len(dps), 2)
        dps.remove_blanks()
        self.assertEqual(len(dps), 1)
    
    def test_collapse(self):
        dps = DraftingPens([
            DraftingPens([DraftingPens([DraftingPen()])]),
            DraftingPens([DraftingPen()]),
        ])

        dps.collapse() # should not mutate by default
        self.assertIsInstance(dps[0], DraftingPens)
        self.assertIsInstance(dps[0][0], DraftingPens)

        dps.collapse(onself=True) # now it should mutate
        self.assertEqual(len(dps), 2)
        self.assertNotIsInstance(dps[0], DraftingPens)
        self.assertNotIsInstance(dps[1], DraftingPens)
    
    def test_find(self):
        dps = DraftingPens([
            DraftingPens([DraftingPens([DraftingPen().tag("find-me").f(hsl(0.9))])]),
            DraftingPen().tag("not-me"),
            DraftingPens([DraftingPen().tag("find-me").f(hsl(0.3))])])

        self.assertEqual(dps.find("find-me")[0].f().h/360, 0.9)
        self.assertAlmostEqual(dps.find("find-me")[1].f().h/360, 0.3)

    def test_cond(self):
        dps = (DraftingPens([
            (DraftingPen().cond(True,
                lambda p: p.f(rgb(1, 0, 0))))]))
        
        self.assertEqual(dps[0].f().r, 1)

        def _build(condition):
            return (DraftingPens([
                (DraftingPen().cond(condition,
                    lambda p: p.f(rgb(0, 0, 1)),
                    lambda p: p.f(rgb(1, 0, 0))))]))
        
        self.assertEqual(_build(True)[0].f().b, 1)
        self.assertEqual(_build(False)[0].f().r, 1)
    
    def test_alpha(self):
        dps = (DraftingPens([
            (DraftingPens([
                (DraftingPen().a(0.5))
            ]).a(0.5))
        ]).a(0.25))

        def walker(p, pos, data):
            if pos == 0:
                self.assertEqual(data["alpha"], 0.0625)
            elif pos == 1 and data["depth"] == 0:
                self.assertEqual(data["alpha"], 0.25)
            elif pos == 1 and data["depth"] == 1:
                self.assertEqual(data["alpha"], 0.125)

        dps.walk(walker)
    
    def test_visibility(self):
        dps = (DraftingPens([
            (DraftingPens([
                (DraftingPen().v(1).tag("visible")),
                (DraftingPen().v(0).tag("invisible"))
            ]))
        ]))

        def walker(p, pos, data):
            nonlocal visible_pen_count
            if pos == 0:
                visible_pen_count += 1

        visible_pen_count = 0
        dps.walk(walker, visible_only=True)
        self.assertEqual(visible_pen_count, 1)

        visible_pen_count = 0
        dps.walk(walker, visible_only=False)
        self.assertEqual(visible_pen_count, 2)

        visible_pen_count = 0
        dps[0][0].v(0)
        dps.walk(walker, visible_only=True)
        self.assertEqual(visible_pen_count, 0)
    
    def test_style(self):
        src = """
from coldtype import *

def two_styles(r):
    return (DATPen()
        .oval(r.inset(50).square())
        .f(hsl(0.8))
        .attr("alt", fill=hsl(0.3)))

@renderable()
def no_style_set(r):
    return two_styles(r)

@renderable(style="alt")
def style_set(r):
    return two_styles(r)

def lattr_styles(r):
    return (DATPen()
        .oval(r.inset(50).square())
        .f(hsl(0.5)).s(hsl(0.7)).sw(5)
        .lattr("alt", lambda p: p.f(hsl(0.7)).s(hsl(0.5)).sw(15)))

@renderable()
def lattr_no_style(r):
    return lattr_styles(r)

@renderable(style="alt")
def lattr_style_set(r):
    return lattr_styles(r)
"""

        sr = SourceReader(None, code=src)
        rs = sr.frame_results(0)
        sr.unlink()

        self.assertNotEqual(
            rs[0][-1][0].attr(rs[0][0].style, "fill"),
            rs[1][-1][0].attr(rs[1][0].style, "fill"))
        
        self.assertNotEqual(
            rs[2][-1][0].attr(rs[2][0].style, "fill"),
            rs[3][-1][0].attr(rs[3][0].style, "fill"))
        
        self.assertEqual(rs[2][-1][0].attr(rs[2][0].style, "stroke").get("weight"), 5)
        self.assertEqual(rs[3][-1][0].attr(rs[3][0].style, "stroke").get("weight"), 15)

        dpm = DrawablePenMixin()
        dpm.dat = rs[3][-1][0]
        attrs = [x for _, x in list(dpm.findStyledAttrs(rs[3][0].style))]
        self.assertEqual(len(attrs), 2)
        self.assertEqual(attrs[1][1].get("weight"), 15)

    def test_subsegmenting(self):
        f1 = Font.Cacheable("assets/ColdtypeObviously_BlackItalic.ufo")

        shape = (StSt("C", f1, 1000, wght=0.5)[0]
            .explode()[0])

        self.assertAlmostEqual(
            shape.length()/2,
            shape.copy().subsegment(0, 0.5).length(),
            delta=1)
        
        self.assertAlmostEqual(
            shape.length(),
            shape.copy().subsegment(0, 1).length(),
            delta=1)
        
        shape1 = (StSt("D", f1, 1000, wght=0.5)[0]
            .explode()[0])
        
        shape2 = shape1.copy().fully_close_path()

        self.assertLess(shape1.length(), shape2.length())

        self.assertAlmostEqual(
            shape2.length()/2,
            shape2.copy().subsegment(0, 0.5).length(),
            delta=1)
    
    def test_explode(self):
        r = Rect(1000, 500)
        
        o1 = (StSt("O", co, 500, wdth=1).pen())
        o2 = (StSt("O", co, 500, wdth=1)
            .pen()
            .explode()
            .index(1, lambda p: p.rotate(90))
            .implode().f(hsl(0.3)).align(r))
        
        self.assertEqual(
            o1.explode()[0].ambit().w,
            o2.explode()[0].ambit().w)

        self.assertAlmostEqual(
            o1.explode()[1].ambit().h,
            o2.explode()[1].ambit().w)
    
    def test_scaleToRect(self):
        r = Rect(1000, 500)
        dps = DraftingPens([
            (StSt("SPACEFILLING", mutator, 50)
                .align(r)
                .f(hsl(0.8))
                .scaleToRect(r.inset(100, 100), False)),
            (StSt("SPACEFILLING", mutator, 50)
                .align(r)
                .f(hsl(0.5))
                .scaleToWidth(r.w-20)),
            (StSt("SPACEFILLING", mutator, 50)
                .align(r)
                .f(hsl(0.3))
                .scaleToHeight(r.h-50))])
        
        self.assertAlmostEqual(
            dps[0].ambit(th=1).w, r.inset(100).w)
        self.assertAlmostEqual(
            dps[0].ambit(tv=1).h, r.inset(100).h)
        self.assertAlmostEqual(
            dps[1].ambit(th=1).w, r.w-20)
        self.assertAlmostEqual(
            dps[2].ambit(tv=1).h, r.h-50)
        
        dps.picklejar(r)
    
    def test_distribute_and_track(self):
        dps = DraftingPens()
        rnd = Random(0)
        r = Rect(1000, 500)

        for _ in range(0, 11):
            dps += (DraftingPen()
                .rect(Rect(100, 100))
                .f(hsl(rnd.random(), s=0.6))
                .rotate(rnd.randint(-45, 45)))
        dps = (dps
            .distribute()
            .track(-50)
            .reversePens()
            .understroke(s=0.2).align(r))
        
        self.assertEqual(len(dps), 22)
        self.assertEqual(dps.ambit(th=1).round().w, 830)
    
        dps.picklejar(r)
    
    def test_track_to_rect(self):
        r = Rect(1000, 500)
        text = (StSt("COLD", co, 300, wdth=0, r=1)
            .align(r)
            .track_to_rect(r.inset(50, 0), r=1))
        
        self.assertEqual(text[0].glyphName, "D")
        self.assertEqual(text[-1].ambit().round().x, 50)
        self.assertEqual(text[0].ambit().round().x, 883)
    
    def test_sample(self):
        r = Rect(1000, 500)
        dp = (DraftingPen()
            .define(r=r)
            .gs("$r↖ $r↗ $r↙|↘|65 ɜ"))
        # dp = (DraftingPen()
        #     .define(r=r)
        #     .gs("$r↙ $r↗ ɜ")
        #     .fssw(None, 0, 5))
        dp.picklejar(r)


if __name__ == "__main__":
    unittest.main()


import unicodedata, unittest
from pathlib import Path
from coldtype.geometry import Rect
from coldtype.helpers import glyph_to_uni
from coldtype.pens.datpen import DATPen, DATPens
from coldtype.text.composer import StSt, Font, Style, Slug, Graf, SegmentedString

tf = Path(__file__).parent
co = Font.Cacheable("assets/ColdtypeObviously-VF.ttf")
mutator = Font.Cacheable("assets/MutatorSans.ttf")
zh = Font.Cacheable("assets/NotoSansCJKsc-Black.otf")

latin_font = Font("assets/NotoSans-Black.ttf")
arabic_font = Font("~/Type/fonts/fonts/_i18n/GretaArabicCompressedAR-Heavy.otf")
ar_light = Font("~/Type/fonts/fonts/_i18n/GretaArabicCondensedAR-Light.otf")
hebrew_font = Font("~/Type/fonts/fonts/_i18n/GretaSansCondensedH+L-Medium.otf")

r = Rect(1000, 500)

def gn_to_c(gn):
    return chr(int(gn.replace("uni", ""), 16))

def gn_to_uniname(gn):
    return unicodedata.name(gn_to_c(gn))

class TestI18N(unittest.TestCase):
    def test_mixed_lang_slug(self):
        obv = Style(co, 300, wdth=1, wght=0)
        style = Style(zh, 300, lang="zh")
        dps = (Slug("CO同步", style, obv)
            .fit(r.w-100)
            .pens()
            .align(r))
        
        self.assertEqual(
            dps[-1].ambit().round(),
            Rect([657,138,300,220]))
        
        self.assertEqual(dps[0].glyphName, "C")
        self.assertEqual(dps[1].glyphName, "O")
        self.assertEqual(dps[2].glyphName, "cid12378")
        self.assertEqual(dps[3].glyphName, "cid23039")
    
    def test_rtl_multiline(self):
        ar = 'Limmmm/Satلل\nوصل الإستيرِو'
        lines = ar.split("\n")
        latin = Style(latin_font, 100, fill=("hr", 0.5, 0.5))
        arabic = Style(ar_light, 150, lang="ar", bs=-1)
        slugs = [
            Slug(lines[0], arabic, latin),
            Slug(lines[1], arabic, latin)
        ]
        graf = Graf(slugs, r, leading=30)
        dps = graf.pens().xa().align(r)

        lgn = dps[-1][-1].glyphName
        lc = gn_to_c(lgn)
        
        self.assertEqual(
            unicodedata.name(lc),
            "ARABIC LETTER WAW")
        self.assertEqual(
            dps[-1][-1].ambit().round(),
            Rect([707,148,42,87]))
        
        self.assertEqual(dps[0][0].glyphName, "L")
        self.assertEqual(dps[0][-1].glyphName, "uniFEDF")
    
    def test_hebrew(self):
        hebrew = Style(hebrew_font, 130)
        slug = Slug('קומפרסיה ועוד', hebrew)
        dps = slug.pens().align(r)

        self.assertEqual(gn_to_c(dps[-1].glyphName), "ק")
        self.assertEqual(dps[-1].ambit().round(),
            Rect([719,212,52,76]))
    
    def test_multidir_seg_string(self):
        txt = "+بوابة"
    
        latin = Style(latin_font, 130, fill=("hr", 0.5, 0.5))
        arabic = Style(arabic_font, 150, lang="ar", bs=-1)
        seg = SegmentedString(txt, dict(Arab=arabic, Latn=latin)).pens()
        slug = Slug(txt, arabic, latin).pens()
        dps = DATPens([
            seg.align(r).translate(0, 100),
            slug.align(r).translate(0, -100)])
        
        self.assertEqual(dps[0][-1].glyphName, "plus")
        self.assertEqual(
            gn_to_uniname(dps[0][0].glyphName),
            "ARABIC LETTER TEH MARBUTA FINAL FORM")
        
        self.assertEqual(dps[1][0].glyphName, "plus")
        self.assertEqual(
            gn_to_uniname(dps[1][1].glyphName),
            "ARABIC LETTER TEH MARBUTA FINAL FORM")
        
    def test_combine_slugs(self):
        s1 = Slug("YO", Style(co, 300, wdth=1)).pens()
        line = DATPen().rect(Rect(100, 20))
        s2 = Slug("OY", Style(co, 300, wdth=0)).pens()
        shape = DATPen().oval(Rect(100, 100))
        dps = DATPens([s1, line, s2, shape]).distribute().align(r)
        self.assertEqual(dps.ambit().round(),
            Rect([127,138,737,225]))
        self.assertEqual(dps[1].ambit().round(),
            Rect([540,138,100,20]))
    
    def test_language_specific_forms(self):
        fnt = "assets/SourceSerifPro-Black.ttf"
        txt = StSt("ríjks", fnt, 350, lang="NLD")
        self.assertEqual(txt[1].glyphName, "iacute")
        self.assertEqual(txt[2].glyphName, "uni0237")
        txt = StSt("ríjks", fnt, 350)
        self.assertEqual(txt[1].glyphName, "iacute")
        self.assertEqual(txt[2].glyphName, "j")
    
    def test_language_specific_ufo(self):
        hershey_gothic = Font.Cacheable("~/Type/fonts/fonts/hershey/Hershey-TriplexGothicGerman.ufo")
        txt = StSt("Grieß".upper(), hershey_gothic, 200, tu=-100)
        self.assertEqual(len(txt), 6)
        self.assertEqual(txt[-1].glyphName, "S")
        self.assertEqual(txt[-2].glyphName, "S")


if __name__ == "__main__":
    unittest.main()


import os
from coldtype import *

if False:
    r = Rect(1000, 300)
    DP().oval(Rect(200, 200)).align(r).f(hsl(0.9)).picklejar(r)

    r = Rect(500, 300)
    DP().oval(Rect(200, 200)).align(r).f(hsl(0.3)).picklejar(r)

picklejar = Path("~/.coldtype/picklejar").expanduser()
if picklejar.exists():
    pickles = list(picklejar.glob("*.pickle"))
    pickles.sort(key=os.path.getmtime)
else:
    picklejar = None
    pickles = []

results = []
for pickle in pickles:
    results.append([
        pickle,
        DP().Unpickle(pickle)])

def make_renderable(name, result):
    r = result.data.get("rect", Rect(1000, 1000))
    @renderable(r)
    def pj(r): # TODO naming needs to be different?
        return result
    pj.name = name
    return pj

RENDERABLES = []
for result in results:
    pickle, res = result
    try:
        RENDERABLES.append(make_renderable(pickle.stem, res))
    except:
        print("failed to load", pickle.stem)

for pickle, _ in results:
    pickle.unlink()


from coldtype.renderable.renderable import Overlay, Action, RenderPass, renderable, skia_direct, svgicon, glyph, fontpreview, iconset, SkiaPen
from coldtype.renderable.animation import animation, aframe, FFMPEGExport


import json
from pathlib import Path
from coldtype.geometry import Point


class RendererStateEncoder(json.JSONEncoder):
    def default(self, o):
        return {
            "controller_values": o.controller_values
        }


class RendererState():
    def __init__(self, renderer):
        self.renderer = renderer
        self.previewing = False
        self.preview_scale = 1
        self.controller_values = {}
        self.mouse = Point(0, 0)
        self.overlays = {}
        self._frame_offsets = {}
        self._initial_frame_offsets = {}
        self.canvas = None
        self._last_filepath = None
        self.cv2caps = {}
        self.reset()
    
    def reset(self, ignore_current_state=False):
        if self.filepath == self._last_filepath and not ignore_current_state:
            return
        
        if self.filepath:
            self._last_filepath = self.filepath
            try:
                deserial = json.loads(self.filepath.read_text())
                cv = deserial.get("controller_values")
                if cv:
                    self.controller_values = cv
            except json.decoder.JSONDecodeError:
                self.controller_values = {}
            except FileNotFoundError:
                self.controller_values = {}
    
    def clear(self):
        if self.filepath:
            self.filepath.write_text("")
        self.reset()
    
    @property
    def filepath(self):
        if self.renderer.source_reader.filepath:
            return Path(str(self.renderer.source_reader.filepath).replace(".py", "") + "_state.json")
        else:
            return None
    
    @property
    def midi(self):
        return self.controller_values
    
    def persist(self):
        if self.filepath:
            print("Saving Controller State...")
            self.filepath.write_text(RendererStateEncoder().encode(self))
        else:
            print("No source; cannot persist state")
    
    def record_mouse(self, pos):
        self.mouse = pos.scale(1/self.preview_scale)
        #return Action.PreviewStoryboard
    
    def on_mouse_button(self, pos, btn, action, mods):
        return self.record_mouse(pos)
    
    def on_mouse_move(self, pos):
        return self.record_mouse(pos)
    
    def mod_preview_scale(self, inc, absolute=0):
        if absolute > 0:
            ps = absolute
        else:
            ps = self.preview_scale + inc
        self.preview_scale = max(0.1, min(5, ps))
    
    def add_frame_offset(self, key, offset):
        if key in self._frame_offsets:
            offsets = self._frame_offsets[key]
            initials = self._initial_frame_offsets[key]
            offsets.append(offset)
            initials.append(offset)
        else:
            self._frame_offsets[key] = [offset]
            self._initial_frame_offsets[key] = [offset]
    
    def get_frame_offsets(self, key):
        return self._frame_offsets.get(key, [0])
    
    def adjust_all_frame_offsets(self, adj, absolute=False):
        if absolute:
            for k, v in self._frame_offsets.items():
                for i, o in enumerate(v):
                    v[i] = adj
        else:
            for k, v in self._frame_offsets.items():
                for i, o in enumerate(v):
                    v[i] = o + adj
    
    def adjust_keyed_frame_offsets(self, key, fn):
        offs = self.get_frame_offsets(key)
        for i, o in enumerate(offs):
            offs[i] = fn(i, o)
    
    def toggle_overlay(self, overlay):
        v = not self.overlays.get(overlay, False)
        if not v:
            if overlay in self.overlays:
                del self.overlays[overlay]
        else:
            self.overlays[overlay] = True


import subprocess

BLENDER = "/Applications/Blender.app/Contents/MacOS/blender"

def blender_launch_livecode(file):
    #call = f"{BLENDER} {file}"
    print(f"Opening blend file: {file}...")
    return subprocess.Popen([BLENDER, file, "--python-expr", "from coldtype.blender.watch import watch; watch()"])


def blend_frame(py_file, blend_file, expr, output_dir, fi):
    call = f"{BLENDER} -b \"{blend_file}\" --python-expr \"{expr}\" -o \"{output_dir}####.png\" -f {fi}"
    print(f"Blending frame {fi}...")
    print(call)
    #return
    #os.system(call)
    process = subprocess.Popen(call, stdout=subprocess.PIPE, shell=True)
    log = ""
    while True:
        out = process.stdout.read(1).decode("utf-8")
        log += out
        if out == "" and process.poll() != None:
            break
        if "Error: Python:" in log:
            print(log)
            process.kill()
            process.terminate()
            break
    print(log)
    print(f"/Blended frame {fi}.")

def blend_source(py_file, blend_file, frame, output_dir, samples=2, denoise=True):
    """
    A facility for telling Blender to render a single frame in a background process
    """
    expr = f"from coldtype.blender.render import frame_render; frame_render('{py_file}', {frame}, {samples}, {denoise})"
    #print(expr)
    blend_frame(py_file, blend_file, expr, output_dir, frame)

def frame_render(file, frame, samples, denoise=True):
    """
    A facility for easy-rendering from within a backgrounded blender
    """
    import bpy
    from coldtype.renderer.reader import SourceReader
    from coldtype.blender import walk_to_b3d
    bpy.data.scenes[0].cycles.samples = samples
    if denoise:
        bpy.data.scenes[0].cycles.denoiser = "OPENIMAGEDENOISE"
        bpy.data.scenes[0].cycles.use_denoising = True
    else:
        bpy.data.scenes[0].cycles.use_denoising = False

    sr = SourceReader(file)
    for r, res in sr.frame_results(frame, class_filters=[r"^b3d_.*$"]):
        walk_to_b3d(res, dn=True, renderable=r)
    sr.unlink()


# to be loaded from within Blender

import os, math
from pathlib import Path

from coldtype.geometry.rect import Rect
from coldtype.pens.datpen import DATPen, DATPens
from coldtype.pens.blenderpen import BlenderPen, BPH
from coldtype.color import hsl

from coldtype.time import Frame, Timeline
from coldtype.renderable import renderable, Overlay, Action
from coldtype.renderable.animation import animation

from coldtype.blender.render import blend_source

try:
    import bpy # noqa
except ImportError:
    bpy = None
    pass

def b3d(collection,
    callback=None,
    plane=False,
    dn=False,
    material=None,
    zero=False,
    upright=False,
    tag_prefix=None,
    ):
    if not bpy: # short-circuit if this is meaningless
        return lambda x: x

    if not isinstance(collection, str):
        callback = collection
        collection = "Coldtype"

    pen_mod = None
    if callback and not callable(callback):
        pen_mod = callback[0]
        callback = callback[1]

    def annotate(pen:DATPen):
        if bpy and pen_mod:
            pen_mod(pen)
        
        prev = pen.data.get("b3d", {})
        if prev:
            callbacks = [*prev.get("callbacks"), callback]
        else:
            callbacks = [callback]

        #c = None
        #if zero:
        #    c = pen.ambit().pc
        #    pen.translate(-c.x, -c.y)

        pen.add_data("b3d", dict(
            collection=(collection
                or prev.get("collection", "Coldtype")),
            callbacks=callbacks,
            material=(material
                or prev.get("material", "ColdtypeDefault")),
            tag_prefix=(tag_prefix or prev.get("tag_prefix")),
            dn=dn,
            plane=plane,
            zero=zero,
            #reposition=c,
            upright=upright))
    
    return annotate


def b3d_post(callback):
    if not bpy: # short-circuit for non-bpy
        return lambda x: x

    def _b3d_post(pen:DATPen):
        prev = pen.data.get("b3d_post")
        if prev:
            callbacks = [*prev, callback]
        else:
            callbacks = [callback]
        pen.data["b3d_post"] = callbacks
    
    return _b3d_post


def b3d_pre(callback):
    def _cast(pen:DATPen):
        if bpy:
            callback(pen)
    return _cast


def walk_to_b3d(result:DATPens,
    dn=False,
    renderable=None,
    ):
    built = {}

    center = renderable.center
    center_rect = renderable.rect

    def walker(p:DATPen, pos, data):
        if pos == 0:
            bdata = p.data.get("b3d")
            if not bdata:
                p.ch(b3d(lambda bp: bp.extrude(0.01)))
                bdata = p.data.get("b3d")
            
            zero = bdata.get("zero", False)

            if center and True:
                cx = -center_rect.w/2*(1-center[0])
                cy = -center_rect.h/2*(1-center[1])
                p.translate(cx, cy)

            pc = p.ambit().pc
            if zero:
                p.translate(-pc.x, -pc.y)
            
            if p.tag() == "?" and data.get("idx"):
                tag = "_".join([str(i) for i in data["idx"]])
                if bdata.get("tag_prefix"):
                    tag = bdata.get("tag_prefix") + tag
                else:
                    tag = "ct_autotag_" + tag
                p.tag(tag)

            if bdata:
                coll = BPH.Collection(bdata["collection"])
                material = bdata.get("material", "ColdtypeDefault")

                if len(p.value) == 0:
                    p.v(0)
                
                denovo = bdata.get("dn", dn)

                if bdata.get("plane"):
                    bp = p.cast(BlenderPen).draw(coll, plane=True, material=material, dn=True)
                else:
                    bp = p.cast(BlenderPen).draw(coll, dn=denovo, material=material)
                
                bp.rotate(0)
                
                if bdata.get("callbacks"):
                    for cb in bdata.get("callbacks"):
                        cb(bp)

                bp.hide(not p._visible)

                # if center and False:
                #     cx = -center_rect.w/2*(1-center[0])
                #     cy = -center_rect.h/2*(1-center[1])
                #     bp.locate_relative(cx/100, cy/100)

                if renderable:
                    if renderable.upright:
                        bp.rotate(90)

                if zero: #bdata.get("reposition"):
                    pt = pc #bdata.get("reposition")
                    if bdata.get("upright"):
                        bp.locate_relative(pt.x/100, 0, pt.y/100)
                    else:
                        bp.locate_relative(pt.x/100, pt.y/100)
                
                built[p.tag()] = (p, bp)

                b3d_post = p.data.get("b3d_post")
                if b3d_post:
                    for post in b3d_post:
                        post(bp)
                
    result.walk(walker)

class b3d_renderable(renderable):
    def __init__(self,
        rect=(1080, 1080),
        center=(0, 0),
        upright=False,
        **kwargs
        ):
        self.center = center
        self.upright = upright
        super().__init__(rect, **kwargs)

    def post_read(self):
        if not hasattr(self, "blend"):
            self.blend = self.filepath.parent / "blends" / (self.filepath.stem + ".blend")

        if self.blend:
            self.blend = Path(self.blend).expanduser()
            self.blend.parent.mkdir(exist_ok=True, parents=True)

        super().post_read()


class b3d_animation(animation):
    def __init__(self,
        rect=(1080, 1080),
        samples=16,
        denoise=True,
        blend=None,
        match_length=True,
        match_output=True,
        bake=False,
        center=(0, 0),
        upright=False,
        **kwargs
        ):
        self.func = None
        self.name = None
        self.current_frame = -1
        self.samples = samples
        self.denoise = denoise
        self.blend = blend
        self.bake = bake
        self.center = center
        self.upright = upright
        self.match_length = match_length
        self.match_output = match_output
        
        if "timeline" not in kwargs:
            kwargs["timeline"] = Timeline(30)
        
        super().__init__(rect=rect, **kwargs)

        if bpy and self.match_length:
            bpy.data.scenes[0].frame_end = self.t.duration-1
            # don't think this is totally accurate but good enough for now
            if isinstance(self.t.fps, float):
                bpy.data.scenes[0].render.fps = round(self.t.fps)
                bpy.data.scenes[0].render.fps_base = 1.001
            else:
                bpy.data.scenes[0].render.fps = self.t.fps
                bpy.data.scenes[0].render.fps_base = 1
            
    def frame_img(self, fi):
        from coldtype.img.skiaimage import SkiaImage
        return SkiaImage(self.pass_path(fi))
    
    def run(self, render_pass, renderer_state):
        fi = render_pass.args[0].i
        if renderer_state and not bpy:
            if renderer_state.previewing:
                if Overlay.Rendered in renderer_state.overlays:
                    return self.frame_img(fi)
        
        return super().run(render_pass, renderer_state)
    
    def rasterize(self, _, rp):
        fi = rp.args[0].i
        blend_source(self.filepath, self.blend, fi, self.pass_path(""), self.samples, denoise=self.denoise)
        return True
    
    def post_read(self):
        if not self.blend:
            self.blend = self.filepath.parent / "blends" / (self.filepath.stem + ".blend")

        if self.blend:
            self.blend = Path(self.blend).expanduser()
            self.blend.parent.mkdir(exist_ok=True, parents=True)

        super().post_read()
        if bpy and self.match_output:
            bpy.data.scenes[0].render.filepath = str(self.pass_path(""))
    
    def baked_frames(self):
        def bakewalk(p, pos, data):
            if pos == 0:
                fi = data["idx"][0]
                (p.ch(b3d(f"CTBakedAnimation_{self.name}",
                    lambda bp: bp
                        .show_on_frame(fi)
                        .print(f"Baking frame {fi}..."),
                    dn=True,
                    tag_prefix=f"ct_baked_frame_{fi}_{self.name}")))
        
        to_bake = DATPens([])
        for ps in self.passes(Action.RenderAll, None)[:]:
            to_bake += self.run_normal(ps, None)
        
        return to_bake.walk(bakewalk)


import contextlib
import drawBot as db
from coldtype.geometry import Point, Line, Rect
from coldtype.pens.drawbotpen import DrawBotPen
from coldtype.pens.draftingpen import DraftingPen
from coldtype.pens.draftingpens import DraftingPens
from coldtype.pens.datpen import DATPen
from coldtype.text.reader import StyledString, Style, Font
from coldtype.text.composer import StSt
from coldtype.color import hsl, bw
from coldtype.time import Frame
from pathlib import Path

from coldtype.renderer import renderable, Action
from coldtype.renderable.animation import animation, RenderPass

try:
    import drawBot as db
    import AppKit
except ImportError:
    db = None

class drawbot_script(renderable):
    def __init__(self, rect=(1080, 1080), scale=1, **kwargs):
        if not db:
            raise Exception("DrawBot not installed!")
        super().__init__(rect=Rect(rect).scale(scale), rasterizer="drawbot", **kwargs)
        self.self_rasterizing = True
    
    def normalize_result(self, pens):
        return pens
    
    def run(self, render_pass, renderer_state):
        from coldtype.pens.drawbotpen import DrawBotPen
        use_pool = True
        if use_pool:
            pool = AppKit.NSAutoreleasePool.alloc().init()
        try:
            db.newDrawing()
            if renderer_state.previewing:
                ps = renderer_state.preview_scale
                db.size(self.rect.w*ps, self.rect.h*ps)
                db.scale(ps, ps)
                if not renderer_state.renderer.source_reader.config.window_transparent:
                    DATPen().rect(self.rect).f(self.bg).cast(DrawBotPen).draw()
            else:
                db.size(self.rect.w, self.rect.h)
            if self.rstate:
                render_pass.fn(*render_pass.args, renderer_state)
            else:
                render_pass.fn(*render_pass.args)
            result = None
            if renderer_state.previewing:
                previews = (render_pass.output_path.parent / "_previews")
                previews.mkdir(exist_ok=True, parents=True)
                preview_frame = previews / render_pass.output_path.name
                db.saveImage(str(preview_frame))
                result = preview_frame
            else:
                render_pass.output_path.parent.mkdir(exist_ok=True, parents=True)
                db.saveImage(str(render_pass.output_path))
                result = render_pass.output_path
            db.endDrawing()
        finally:
            if use_pool:
                del pool
        return result

class drawbot_animation(drawbot_script, animation):
    def passes(self, action, renderer_state, indices=[]):
        if action in [
            Action.RenderAll,
            Action.RenderIndices,
            Action.RenderWorkarea]:
            frames = super().active_frames(action, renderer_state, indices)
            passes = []
            for i in frames:
                p = RenderPass(self, action, "{:04d}".format(i), [Frame(i, self)])
                passes.append(p)
            return passes
        else:
            return super().passes(action, renderer_state, indices)

def dbdraw(p:DraftingPen):
    p.cast(DrawBotPen).draw()
    return p

def tobp(p:DraftingPen):
    bp = db.BezierPath()
    p.replay(bp)
    return bp

def dbdraw_with_filters(rect:Rect, filters):
    def _draw_call(p:DraftingPen):
        p.cast(DrawBotPen).draw_with_filters(rect, filters)
        return p
    return _draw_call

def page_rect() -> Rect:
    return Rect(db.width(), db.height())

@contextlib.contextmanager
def new_page(r:Rect=Rect(1000, 1000)):
    _r = Rect(r)
    db.newPage(*_r.wh())
    yield _r

@contextlib.contextmanager
def new_drawing(rect:Rect=Rect(1000, 1000), count=1, save_to=None):
    db.newDrawing()
    for idx in range(0, count):
        with new_page(rect) as r:
            yield idx, r
    if save_to:
        db.saveImage(str(save_to))
    db.endDrawing()

def pdfdoc(fn, path, frame_class=Frame):
    db.newDrawing()
    r = fn.rect
    w, h = r.wh()
    for idx in range(0, fn.duration):
        print(f"Saving page {idx}...")
        db.newPage(w, h)
        if frame_class:
            fn.func(frame_class(idx, fn))
        else:
            fn.func(r)
    pdf_path = Path(path)
    pdf_path.parent.mkdir(exist_ok=True)
    db.saveImage(str(pdf_path))
    print("Saved pdf", str(pdf_path))
    db.endDrawing()


import bpy
from pathlib import Path
from runpy import run_path
import traceback

from coldtype.renderer.reader import SourceReader
from coldtype.blender import b3d_animation, walk_to_b3d

#from coldtype.blender.watch import watch; watch()

# original idea: https://blender.stackexchange.com/questions/15670/send-instructions-to-blender-from-external-application

class ColdtypeWatchingOperator(bpy.types.Operator):
    bl_idname = "wm.coldtype_watching_operator"
    bl_label = "Coldtype Watching Operator"

    _timer = None
    file = Path("~/.coldtype/blender.txt").expanduser()
    sr = None
    current_frame = -1

    def render_current_frame(self, statics=False):
        animation_found = False

        cfs = [r"^b3d_.*$"]
        if not statics:
            cfs = [r"^b3d_animation$"]
        
        for r, res in self.sr.frame_results(
            self.current_frame,
            class_filters=cfs
            ):

            if isinstance(r, b3d_animation):
                if r.bake:
                    if statics:
                        walk_to_b3d(r.baked_frames(), renderable=r)
                else:
                    animation_found = True
                    walk_to_b3d(res, renderable=r)
            elif statics:
                walk_to_b3d(res, renderable=r)
        
        if not animation_found:
            bpy.data.scenes[0].frame_set(0)
        return animation_found

    def reimport(self, arg):
        try:
            self.sr = SourceReader(arg)
            self.sr.unlink()
            #bpy.data.scenes[0].frame_start = 0

            bpy.app.handlers.frame_change_pre.clear()

            self.current_frame = bpy.context.scene.frame_current
            animation_found = self.render_current_frame(statics=True)
            
            if not animation_found:
                return

            def _frame_update_handler(scene):
                if scene.frame_current != self.current_frame:
                    self.current_frame = scene.frame_current
                    self.render_current_frame(statics=False)
            
            bpy.app.handlers.frame_change_pre.append(_frame_update_handler)
        
        except Exception as e:
            self.current_frame = -1
            bpy.app.handlers.frame_change_pre.clear()
            stack = traceback.format_exc()
            print("---"*10)
            print(stack)
        
        print(f"ran {arg}")

    def modal(self, context, event):
        if event.type == 'TIMER':
            if not self.file.exists():
                return {'PASS_THROUGH'}
            
            for line in self.file.read_text().splitlines():
                line = line.rstrip("\n")
                cmd, arg = line.split(",")
                if cmd == 'import':
                    self.reimport(arg)
                elif cmd == "play_preview":
                    bpy.ops.screen.animation_play()
                elif cmd == "frame_offset":
                    bpy.ops.screen.frame_offset(delta=int(arg))
                elif cmd == 'cancel':
                    self.cancel( context )
                else:
                    print('unknown request=%s arg=%s' % (cmd,arg))
            
            if self.file.exists():
                self.file.unlink()

        return {'PASS_THROUGH'}

    def execute(self, context):
        wm = context.window_manager
        self._timer = wm.event_timer_add(0.1, window=context.window)
        wm.modal_handler_add(self)

        return {'RUNNING_MODAL'}

    def cancel(self, context):
        wm = context.window_manager
        wm.event_timer_remove(self._timer)
        print('timer removed')

def register_watcher():
    bpy.utils.register_class(ColdtypeWatchingOperator)

def unregister_watcher():
    bpy.utils.unregister_class(ColdtypeWatchingOperator)

def watch():
    register_watcher()
    bpy.ops.wm.coldtype_watching_operator()
    print("...waiting for coldtype...")


from coldtype.renderer.winman.passthrough import WinmanPassthrough

import time as ptime
from subprocess import Popen, PIPE
from pathlib import Path

from coldtype.color import rgb

try:
    import glfw
except ImportError:
    glfw = None

try:
    import skia
    import coldtype.fx.skia as skfx
except ImportError:
    skia = None
    skfx = None

# source: https://github.com/PixarAnimationStudios/USD/issues/1372

def monkeypatch_ctypes():
    import os
    import ctypes.util
    uname = os.uname()
    if uname.sysname == "Darwin" and uname.release >= "20.":
        real_find_library = ctypes.util.find_library
        def find_library(name):
            if name in {"OpenGL", "GLUT"}:  # add more names here if necessary
                return f"/System/Library/Frameworks/{name}.framework/{name}"
            return real_find_library(name)
        ctypes.util.find_library = find_library
    return

try:
    from OpenGL import GL
except ImportError:
    monkeypatch_ctypes()
    try:
        from OpenGL import GL
    except:
        GL = None


from coldtype.geometry import Point, Rect, Edge
from coldtype.renderable import Action
from coldtype.renderer.config import ConfigOption, ColdtypeConfig
from coldtype.renderer.keyboard import KeyboardShortcut, REPEATABLE_SHORTCUTS, shortcuts_keyed

from coldtype.pens.svgpen import SVGPen


class WinmanGLFWSkiaBackground(WinmanPassthrough):
    def __init__(self, config:ColdtypeConfig, renderer):
        # TODO is the glfw stuff here actually necessary?

        if not glfw.init():
            raise RuntimeError('glfw.init() failed')
        
        glfw.window_hint(glfw.VISIBLE, glfw.FALSE)
        glfw.window_hint(glfw.STENCIL_BITS, 8)
        window = glfw.create_window(640, 480, '', None, None)
        glfw.make_context_current(window)

        self.context = skia.GrDirectContext.MakeGL()


class WinmanGLFWSkia():
    def __init__(self, config:ColdtypeConfig, renderer, background=False):
        self.config = config
        self.window = None
        self.background = background
        self.renderer = renderer
        self.primary_monitor = None
        self.last_rect = None
        self.all_shortcuts = shortcuts_keyed()
        self.prev_scale = 0
        self.surface = None

        self.typeface = skia.Typeface.MakeFromFile(str(Path(__file__).parent / "../../demo/RecMono-CasualItalic.ttf"))

        if not glfw.init():
            raise RuntimeError('glfw.init() failed')
        
        glfw.window_hint(glfw.STENCIL_BITS, 8)
        
        if self.config.window_transparent:
            glfw.window_hint(glfw.TRANSPARENT_FRAMEBUFFER, glfw.TRUE)
            glfw.window_hint(glfw.DECORATED, glfw.FALSE)
        else:
            glfw.window_hint(glfw.TRANSPARENT_FRAMEBUFFER, glfw.FALSE)
        
        if self.config.window_passthrough:
            try:
                glfw.window_hint(0x0002000D, glfw.TRUE)
                # glfw.window_hint(glfw.MOUSE_PASSTHROUGH, glfw.TRUE)
            except glfw.GLFWError:
                print("failed to hint window for mouse-passthrough")

        if self.config.window_background:
            glfw.window_hint(glfw.FOCUSED, glfw.FALSE)
        
        if self.config.window_float:
            glfw.window_hint(glfw.FLOATING, glfw.TRUE)

        if not self.background:
            self.window = glfw.create_window(int(10), int(10), '', None, None)
        
        self.window_scrolly = 0
        self.window_focus = 0

        self.find_primary_monitor()

        glfw.make_context_current(self.window)

        glfw.set_key_callback(self.window, self.on_key)
        glfw.set_mouse_button_callback(self.window, self.on_mouse_button)
        glfw.set_cursor_pos_callback(self.window, self.on_mouse_move)
        glfw.set_scroll_callback(self.window, self.on_scroll)
        glfw.set_window_focus_callback(self.window, self.on_focus)

        self.set_window_opacity()

        self.prev_scale = self.get_content_scale()
        self.context = skia.GrDirectContext.MakeGL()
        
        self.copy_previews_to_clipboard = False
    
    def find_primary_monitor(self):
        self.primary_monitor = glfw.get_primary_monitor()
        found_primary = None
        if self.config.monitor_name:
            remn = self.config.monitor_name
            monitors = glfw.get_monitors()
            matches = []
            if remn == "list":
                print("> MONITORS")
            for monitor in monitors:
                mn = glfw.get_monitor_name(monitor)
                if remn == "list":
                    print("    -", mn.decode("utf-8"))
                elif remn in str(mn):
                    matches.append(monitor)
            if len(matches) > 0:
                found_primary = matches[0]
            if found_primary:
                self.primary_monitor = found_primary
    
    def create_surface(self, rect):
        backend_render_target = skia.GrBackendRenderTarget(
            int(rect.w), int(rect.h), 0, 0,
            skia.GrGLFramebufferInfo(0, GL.GL_RGBA8))
        
        self.surface = skia.Surface.MakeFromBackendRenderTarget(
            self.context,
            backend_render_target,
            skia.kBottomLeft_GrSurfaceOrigin,
            skia.kRGBA_8888_ColorType,
            skia.ColorSpace.MakeSRGB())
        
        assert self.surface is not None

    def get_content_scale(self):
        u_scale = self.config.window_content_scale
        
        if u_scale:
            return u_scale
        elif glfw and not self.renderer.args.no_viewer:
            if self.primary_monitor:
                return glfw.get_monitor_content_scale(self.primary_monitor)[0]
            else:
                return glfw.get_window_content_scale(self.window)[0]
        else:
            return 1
    
    def content_scale_changed(self):
        scale_x = self.get_content_scale()
        if scale_x != self.prev_scale:
            self.prev_scale = scale_x
            return True
        return False
    
    def calculate_window_size(self, previews):
        rects = []
        dscale = self.renderer.state.preview_scale
        w = 0
        llh = -1
        lh = -1
        h = 0
        for render, result, rp in previews:
            if hasattr(render, "show_error"):
                sr = render.rect
            else:
                sr = render.rect.scale(dscale, "mnx", "mny").round()
            w = max(sr.w, w)
            if render.layer:
                rects.append(Rect(0, llh, sr.w, sr.h))
            else:
                rects.append(Rect(0, lh+1, sr.w, sr.h))
                llh = lh+1
                lh += sr.h + 1
                h += sr.h + 1
        h -= 1

        frect = Rect(0, 0, w, h)

        if not self.last_rect:
            needs_new_context = True
        else:
            needs_new_context = self.last_rect != frect
        
        self.last_rect = frect
        return frect, rects, dscale, needs_new_context
    
    def update_window(self, frect):
        m_scale = self.get_content_scale()
        scale_x, scale_y = m_scale, m_scale

        ww = int(frect.w/scale_x)
        wh = int(frect.h/scale_y)
        glfw.set_window_size(self.window, ww, wh)

        pin = self.config.window_pin

        if pin:
            work_rect = Rect(glfw.get_monitor_workarea(self.primary_monitor))
            wrz = work_rect.zero()
            edges = Edge.PairFromCompass(pin)
            pinned = wrz.take(ww, edges[0]).take(wh, edges[1]).round()
            if edges[1] == "mdy":
                pinned = pinned.offset(0, -30)
            pinned = pinned.flip(wrz.h)
            pinned = pinned.offset(*work_rect.origin())
            wpi = self.config.window_pin_inset
            pinned = pinned.inset(-wpi[0], wpi[1])
            glfw.set_window_pos(self.window, pinned.x, pinned.y)
        else:
            glfw.set_window_pos(self.window, 0, 0)
    
    def set_title(self, text):
        glfw.set_window_title(self.window, text)

    def set_window_opacity(self, relative=None, absolute=None):
        if relative is not None:
            o = glfw.get_window_opacity(self.window)
            op = o + float(relative)
        elif absolute is not None:
            op = float(absolute)
        else:
            op = float(self.config.window_opacity)
        
        glfw.set_window_opacity(self.window,
            max(0.1, min(1, op)))
    
    def reset(self):
        self.window_scrolly = 0
    
    def should_close(self):
        return glfw.window_should_close(self.window)
    
    def focus(self, force=False):
        if self.window_focus == 0 or force:
            glfw.focus_window(self.window)
            return True
        return False
    
    def allow_mouse(self):
        return True
        
    def on_scroll(self, win, xoff, yoff):
        self.window_scrolly += yoff
        #self.on_action(Action.PreviewStoryboard)
        #print(xoff, yoff)
        #pass # TODO!
    
    def on_focus(self, win, focus):
        self.window_focus = focus
    
    def on_mouse_button(self, _, btn, action, mods):
        if not self.allow_mouse():
            return
        
        pos = Point(glfw.get_cursor_pos(self.window)).scale(2) # TODO should this be preview-scale?
        pos[1] = self.last_rect.h - pos[1]
        requested_action = self.renderer.state.on_mouse_button(pos, btn, action, mods)
        if requested_action:
            self.renderer.action_waiting = requested_action
    
    def on_mouse_move(self, _, xpos, ypos):
        if not self.allow_mouse():
            return
        
        pos = Point((xpos, ypos)).scale(2)
        pos[1] = self.last_rect.h - pos[1]
        requested_action = self.renderer.state.on_mouse_move(pos)
        if requested_action:
            self.renderer.action_waiting = requested_action
    
    def on_key(self, win, key, scan, action, mods):
        self.on_potential_shortcut(key, action, mods)
    
    def repeatable_shortcuts(self):
        return REPEATABLE_SHORTCUTS
    
    def on_potential_shortcut(self, key, action, mods):
        for shortcut, options in self.all_shortcuts.items():
            for modifiers, skey in options:
                if key != skey:
                    continue

                if isinstance(mods, list):
                    mod_matches = mods
                else:
                    mod_matches = [0, 0, 0, 0]
                    for idx, mod in enumerate([glfw.MOD_SUPER, glfw.MOD_ALT, glfw.MOD_SHIFT, glfw.MOD_CONTROL]):
                        if mod in modifiers:
                            if mods & mod:
                                mod_matches[idx] = 1
                        elif mod not in modifiers:
                            if not (mods & mod):
                                mod_matches[idx] = 1
                
                mod_match = all(mod_matches)
                
                if not mod_match and len(modifiers) == 0 and isinstance(mods, list):
                    mod_match = True
                
                if len(modifiers) == 0 and mods != 0 and not isinstance(mods, list):
                    mod_match = False
                
                if mod_match and key == skey:
                    if (action == glfw.REPEAT and shortcut in self.repeatable_shortcuts()) or action == glfw.PRESS:
                        #print(shortcut, modifiers, skey, mod_match)
                        return self.renderer.on_shortcut(shortcut)
    
    def turn_over(self):
        frect, rects, dscale, needs_new_context = self.calculate_window_size(self.renderer.previews_waiting_to_paint)

        if needs_new_context:
            self.create_surface(frect)
            self.update_window(frect)

        GL.glClear(GL.GL_COLOR_BUFFER_BIT)

        did_preview = []

        with self.surface as canvas:
            canvas.clear(skia.Color4f(0.3, 0.3, 0.3, 1))
            if self.config.window_transparent:
                canvas.clear(skia.Color4f(0.3, 0.3, 0.3, 0))
            
            for idx, (render, result, rp) in enumerate(self.renderer.previews_waiting_to_paint):
                rect = rects[idx].offset((frect.w-rects[idx].w)/2, 0).round()

                if self.copy_previews_to_clipboard:
                    try:
                        svg = SVGPen.Composite(result, render.rect, viewBox=render.viewBox)
                        print(svg)
                        process = Popen(
                            'pbcopy', env={'LANG': 'en_US.UTF-8'}, stdin=PIPE)
                        process.communicate(svg.encode('utf-8'))
                    except:
                        print("failed to copy to clipboard")

                try:
                    self.draw_preview(idx, dscale, canvas, rect, (render, result, rp))
                    did_preview.append(rp)
                except Exception as e:
                    short_error = self.renderer.print_error()
                    paint = skia.Paint(AntiAlias=True, Color=skia.ColorRED)
                    canvas.drawString(short_error, 10, 32, skia.Font(None, 36), paint)
            
            self.copy_previews_to_clipboard = False
        
        self.surface.flushAndSubmit()
        glfw.swap_buffers(self.window)
        return did_preview
    
    def poll(self):
        glfw.poll_events()
    
    def show_preloaded_frame(self, path):
        GL.glClear(GL.GL_COLOR_BUFFER_BIT)

        with self.surface as canvas:
            if not self.config.window_transparent:
                c = self.renderer.last_animation.bg
                canvas.clear(c.skia())
            image = skia.Image.MakeFromEncoded(skia.Data.MakeFromFileName(str(path)))
            canvas.drawImage(image, 0, 0)
        
        self.surface.flushAndSubmit()
        glfw.swap_buffers(self.window)

    def draw_preview(self, idx, scale, canvas, rect, waiter):
        if isinstance(waiter[1], Path) or isinstance(waiter[1], str):
            image = skia.Image.MakeFromEncoded(skia.Data.MakeFromFileName(str(waiter[1])))
            if image:
                canvas.drawImage(image, rect.x, rect.y)
            return
        
        render, result, rp = waiter
        error_color = rgb(1, 1, 1).skia()
        canvas.save()
        canvas.translate(0, self.window_scrolly)
        canvas.translate(rect.x, rect.y)
        
        if not self.config.window_transparent:
            canvas.drawRect(skia.Rect(0, 0, rect.w, rect.h), skia.Paint(Color=render.bg.skia()))
        
        if not hasattr(render, "show_error"):
            canvas.scale(scale, scale)
        
        if render.clip:
            canvas.clipRect(skia.Rect(0, 0, rect.w, rect.h))
        
        if render.direct_draw:
            try:
                render.run(rp, self.renderer.state, canvas)
            except Exception as e:
                short_error = self.renderer.print_error()
                render.show_error = short_error
                error_color = rgb(0, 0, 0).skia()
        else:
            if render.composites:
                comp = result.ch(skfx.precompose(render.rect))
                if not self.renderer.last_render_cleared:
                    render.last_result = comp
                else:
                    render.last_result = None
            else:
                comp = result
            
            render.draw_preview(1.0, canvas, render.rect, comp, rp)
        
        if hasattr(render, "show_error"):
            paint = skia.Paint(AntiAlias=True, Color=error_color)
            canvas.drawString(render.show_error, 30, 70, skia.Font(self.typeface, 50), paint)
            canvas.drawString("> See process in terminal for traceback", 30, 120, skia.Font(self.typeface, 32), paint)
        
        canvas.restore()

    def terminate(self):
        glfw.terminate()

        if self.context:
            self.context.abandonContext()